@@ -188,7 +188,13 @@ BETTER_ENUM(
     // Controls number of times a stream gets a write in incremental mode
     WRITES_PER_STREAM = 0x10005,
     // Control connection migration
-    CONNECTION_MIGRATION = 0x10006)
+    CONNECTION_MIGRATION = 0x10006,
+    // Control server-initiated key update interval
+    KEY_UPDATE_INTERVAL = 0x10007,
+    // Controls new stream blocked condition.
+    USE_NEW_STREAM_BLOCKED_CONDITION = 0x10008,
+    // Controls autotune flow control on streams.
+    AUTOTUNE_RECV_STREAM_FLOW_CONTROL = 0x10009)
 
 enum class FrameType : uint64_t {
   PADDING = 0x00,
@@ -292,12 +298,17 @@ enum class LocalErrorCode : uint64_t {
   STREAM_CLOSED = 0x40000002,
   STREAM_NOT_EXISTS = 0x40000003,
   CREATING_EXISTING_STREAM = 0x40000004,
+
+  // Should be used when we're closing the connection locally because of an
+  // error.
+  // If there is no error and connection need to be closed, use
+  // TransportErrorCode::NO_ERROR instead.
   SHUTTING_DOWN = 0x40000005,
+
   RESET_CRYPTO_STREAM = 0x40000006,
   CWND_OVERFLOW = 0x40000007,
   INFLIGHT_BYTES_OVERFLOW = 0x40000008,
   LOST_BYTES_OVERFLOW = 0x40000009,
-  CONGESTION_CONTROL_ERROR = 0x4000000A,
   // This is a retryable error. When encountering this error,
   // the user should retry the request.
   NEW_VERSION_NEGOTIATED = 0x4000000A,
@@ -320,6 +331,7 @@ enum class LocalErrorCode : uint64_t {
   KNOB_FRAME_UNSUPPORTED = 0x4000001B,
   PACER_NOT_AVAILABLE = 0x4000001C,
   RTX_POLICIES_LIMIT_EXCEEDED = 0x4000001D,
+  CONGESTION_CONTROL_ERROR = 0x4000001E,
 };
 
 enum class QuicNodeType : bool {
@@ -335,13 +347,14 @@ enum class QuicVersion : uint32_t {
   MVFST = 0xfaceb002,
   QUIC_V1 = 0x00000001,
   QUIC_V1_ALIAS = 0xfaceb003,
-  // MVFST_EXPERIMENTAL used to set initial congestion window
+  QUIC_V1_ALIAS2 = 0xfaceb004,
+  // MVFST_EXPERIMENTAL enables including and using cwnd hints in 0-rtt session
+  // tickets in QuicServerWorker.cpp
   MVFST_EXPERIMENTAL = 0xfaceb00e, // Experimental alias for MVFST
   MVFST_ALIAS = 0xfaceb010,
   MVFST_INVALID = 0xfaceb00f,
-  // MVFST_EXPERIMENTAL2 enables including and using cwnd hints in 0-rtt session
-  // tickets in QuicServerWorker.cpp
   MVFST_EXPERIMENTAL2 = 0xfaceb011, // Experimental alias for MVFST
+  // MVFST_EXPERIMENTAL3 is used to apply a 2x pace scaling for BBRv2
   MVFST_EXPERIMENTAL3 = 0xfaceb013, // Experimental alias for MVFST
 };
 
@@ -613,6 +626,13 @@ constexpr uint32_t kDefaultMaxDatagramsBuffered = 75;
 // milliseconds
 constexpr std::chrono::milliseconds kMinIntervalBetweenSessionTickets = 100ms;
 
+// Number of packets to write with the current cipher before initiating a key
+// update. This is a conservative number below the confidentiality limit (2^23)
+// derived in the spec for packets of size up to 64k bytes:
+// https://www.rfc-editor.org/rfc/rfc9001.html#name-confidentiality-limit
+constexpr uint64_t kDefaultKeyUpdatePacketCountInterval = 1ull * 1000 * 1000;
+constexpr uint64_t kFirstKeyUpdatePacketCount = 500ull;
+
 enum class ZeroRttSourceTokenMatchingPolicy : uint8_t {
   REJECT_IF_NO_EXACT_MATCH = 0,
   LIMIT_IF_NO_EXACT_MATCH = 1,
@@ -657,6 +677,7 @@ enum class WriteDataReason {
   PATHCHALLENGE,
   PING,
   DATAGRAM,
+  BUFFERED_WRITE,
 };
 
 enum class NoWriteReason {
@@ -716,4 +737,8 @@ constexpr size_t kShortHeaderPaddingModulo = 32;
 constexpr uint8_t kMaxReceivedPktsTimestampsStored = 10;
 constexpr uint8_t kDefaultReceiveTimestampsExponent = 3;
 
+// ECN field values
+constexpr uint8_t kEcnECT1 = 0b01;
+constexpr uint8_t kEcnECT0 = 0b10;
+constexpr uint8_t kEcnCE = 0b11;
 } // namespace quic
