@@ -239,6 +239,7 @@ FrameScheduler::FrameScheduler(
 SchedulingResult FrameScheduler::scheduleFramesForPacket(
     PacketBuilderInterface&& builder,
     uint32_t writableBytes) {
+  size_t shortHeaderPadding = 0;
   builder.encodePacketHeader();
   // We need to keep track of writable bytes after writing header.
   writableBytes = writableBytes > builder.getHeaderBytes()
@@ -322,12 +323,13 @@ SchedulingResult FrameScheduler::scheduleFramesForPacket(
         for (size_t i = 0; i < paddingIncrement; i++) {
           writeFrame(PaddingFrame(), builder);
         }
-        QUIC_STATS(conn_.statsCallback, onShortHeaderPadding, paddingIncrement);
+        shortHeaderPadding = paddingIncrement;
       }
     }
   }
 
-  return SchedulingResult(folly::none, std::move(builder).buildPacket());
+  return SchedulingResult(
+      folly::none, std::move(builder).buildPacket(), shortHeaderPadding);
 }
 
 void FrameScheduler::writeNextAcks(PacketBuilderInterface& builder) {
@@ -607,17 +609,25 @@ folly::Optional<PacketNum> AckScheduler::writeNextAcks(
             .maxReceiveTimestampsPerAck
       : 0;
 
-  // If ack_receive_timestamps are not enabled on *either* end-points OR
-  // the peer requests 0 timestamps, we fall-back to using FrameType::ACK
-  if (!isAckReceiveTimestampsSupported || !peerRequestedTimestampsCount) {
-    ackWriteResult = writeAckFrame(meta, builder, FrameType::ACK);
-  } else {
+  if (conn_.transportSettings.readEcnOnIngress) {
+    // If ECN is enabled and we can use it, this will currently take
+    // priority over sending receive timestamps. There is currently no provision
+    // for a frame time that includes both ECN counts and receive timestamps.
+    // TODO: explore design changes for an ACK frame that supports both ECN and
+    // receive timestamps
+    ackWriteResult = writeAckFrame(meta, builder, FrameType::ACK_ECN);
+  } else if (
+      isAckReceiveTimestampsSupported && (peerRequestedTimestampsCount > 0)) {
+    // Use ACK_RECEIVE_TIMESTAMPS if its enabled on both endpoints AND the peer
+    // requests at least 1 timestamp
     ackWriteResult = writeAckFrameWithReceivedTimestamps(
         meta,
         builder,
         conn_.transportSettings.maybeAckReceiveTimestampsConfigSentToPeer
             .value(),
         peerRequestedTimestampsCount);
+  } else {
+    ackWriteResult = writeAckFrame(meta, builder, FrameType::ACK);
   }
   if (!ackWriteResult) {
     return folly::none;
@@ -801,9 +811,8 @@ bool CryptoStreamScheduler::writeCryptoData(PacketBuilderInterface& builder) {
     if (!res) {
       return cryptoDataWritten;
     }
-    VLOG(4) << "Wrote retransmitted crypto"
-            << " offset=" << buffer.offset << " bytes=" << res->len << " "
-            << conn_;
+    VLOG(4) << "Wrote retransmitted crypto" << " offset=" << buffer.offset
+            << " bytes=" << res->len << " " << conn_;
     cryptoDataWritten = true;
   }
 
@@ -943,7 +952,9 @@ SchedulingResult CloningScheduler::scheduleFramesForPacket(
     auto rebuildResult = rebuilder.rebuildFromPacket(outstandingPacket);
     if (rebuildResult) {
       return SchedulingResult(
-          std::move(rebuildResult), std::move(*internalBuilder).buildPacket());
+          std::move(rebuildResult),
+          std::move(*internalBuilder).buildPacket(),
+          0);
     } else if (
         conn_.transportSettings.dataPathType ==
         DataPathType::ContinuousMemory) {
@@ -962,7 +973,7 @@ SchedulingResult CloningScheduler::scheduleFramesForPacket(
       buf->trimEnd(buf->length() - prevSize);
     }
   }
-  return SchedulingResult(folly::none, folly::none);
+  return SchedulingResult(folly::none, folly::none, 0);
 }
 
 folly::StringPiece CloningScheduler::name() const {
