@@ -10,6 +10,7 @@
 #include <quic/QuicConstants.h>
 #include <quic/QuicException.h>
 #include <quic/api/QuicSocket.h>
+#include <quic/api/QuicTransportFunctions.h>
 #include <quic/common/FunctionLooper.h>
 #include <quic/common/NetworkData.h>
 #include <quic/common/events/QuicEventBase.h>
@@ -35,7 +36,9 @@ enum class CloseState { OPEN, GRACEFUL_CLOSING, CLOSED };
  *    This is needed in order for QUIC to be able to live beyond the lifetime
  *    of the object that holds it to send graceful close messages to the peer.
  */
-class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
+class QuicTransportBase : public QuicSocket,
+                          QuicStreamPrioritiesObserver,
+                          QuicAsyncUDPSocket::WriteCallback {
  public:
   QuicTransportBase(
       std::shared_ptr<QuicEventBase> evb,
@@ -94,9 +97,10 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
 
   folly::Optional<std::string> getAppProtocol() const override;
 
-  void setReceiveWindow(StreamId id, size_t recvWindowSize) override;
+  void setReceiveWindow(StreamId, size_t /*recvWindowSize*/) override {}
 
-  void setSendBuffer(StreamId id, size_t maxUnacked, size_t maxUnsent) override;
+  void setSendBuffer(StreamId, size_t /*maxUnacked*/, size_t /*maxUnsent*/)
+      override {}
 
   uint64_t getConnectionBufferAvailable() const override;
 
@@ -111,6 +115,8 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
   folly::Expected<uint64_t, LocalErrorCode> getMaxWritableOnStream(
       StreamId id) const override;
 
+  [[nodiscard]] uint64_t maxWritableOnConn() const override;
+
   folly::Expected<folly::Unit, LocalErrorCode> setConnectionFlowControlWindow(
       uint64_t windowSize) override;
 
@@ -181,10 +187,10 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
 
   folly::Expected<folly::Unit, LocalErrorCode> notifyPendingWriteOnStream(
       StreamId id,
-      WriteCallback* wcb) override;
+      QuicSocket::WriteCallback* wcb) override;
 
   folly::Expected<folly::Unit, LocalErrorCode> notifyPendingWriteOnConnection(
-      WriteCallback* wcb) override;
+      QuicSocket::WriteCallback* wcb) override;
 
   folly::Expected<folly::Unit, LocalErrorCode> unregisterStreamWriteCallback(
       StreamId id) override;
@@ -522,6 +528,26 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
     QuicTransportBase* transport_;
   };
 
+  class ExcessWriteTimeout : public QuicTimerCallback {
+   public:
+    ~ExcessWriteTimeout() override = default;
+
+    explicit ExcessWriteTimeout(QuicTransportBase* transport)
+        : transport_(transport) {}
+
+    void timeoutExpired() noexcept override {
+      transport_->excessWriteTimeoutExpired();
+    }
+
+    void callbackCanceled() noexcept override {
+      // Do nothing.
+      return;
+    }
+
+   private:
+    QuicTransportBase* transport_;
+  };
+
   class PathValidationTimeout : public QuicTimerCallback {
    public:
     ~PathValidationTimeout() override = default;
@@ -686,6 +712,7 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
  protected:
   void updateCongestionControlSettings(
       const TransportSettings& transportSettings);
+  void updateSocketTosSettings();
   void processCallbacksAfterWriteData();
   void processCallbacksAfterNetworkData();
   void invokeReadDataAndCallbacks();
@@ -741,12 +768,6 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
       bool bidirectional,
       const folly::Optional<StreamGroupId>& streamGroupId = folly::none);
 
-  /**
-   * Helper function - if given error is not set, returns a generic app error.
-   * Used by close() and closeNow().
-   */
-  QuicError maybeSetGenericAppError(folly::Optional<QuicError> error);
-
   /**
    * write data to socket
    *
@@ -777,7 +798,6 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
   void pacedWriteDataToSocket();
 
   uint64_t maxWritableOnStream(const QuicStreamState&) const;
-  uint64_t maxWritableOnConn() const;
 
   void lossTimeoutExpired() noexcept;
   void ackTimeoutExpired() noexcept;
@@ -786,6 +806,7 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
   void keepaliveTimeoutExpired() noexcept;
   void drainTimeoutExpired() noexcept;
   void pingTimeoutExpired() noexcept;
+  void excessWriteTimeoutExpired() noexcept;
 
   void setIdleTimer();
   void scheduleAckTimeout();
@@ -847,8 +868,8 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
   }
 
   bool processCancelCode(const QuicError& cancelCode);
-  void processConnectionSetupCallbacks(const QuicError& cancelCode);
-  void processConnectionCallbacks(const QuicError& cancelCode);
+  void processConnectionSetupCallbacks(QuicError&& cancelCode);
+  void processConnectionCallbacks(QuicError&& cancelCode);
 
   /**
    * The callback function for AsyncUDPSocket to provide the additional cmsgs
@@ -891,8 +912,8 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
   DatagramCallback* datagramCallback_{nullptr};
   PingCallback* pingCallback_{nullptr};
 
-  WriteCallback* connWriteCallback_{nullptr};
-  std::map<StreamId, WriteCallback*> pendingWriteCallbacks_;
+  QuicSocket::WriteCallback* connWriteCallback_{nullptr};
+  std::map<StreamId, QuicSocket::WriteCallback*> pendingWriteCallbacks_;
   CloseState closeState_{CloseState::OPEN};
   bool transportReadyNotified_{false};
   bool handshakeDoneNotified_{false};
@@ -904,6 +925,7 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
   KeepaliveTimeout keepaliveTimeout_;
   DrainTimeout drainTimeout_;
   PingTimeout pingTimeout_;
+  ExcessWriteTimeout excessWriteTimeout_;
   FunctionLooper::Ptr readLooper_;
   FunctionLooper::Ptr peekLooper_;
   FunctionLooper::Ptr writeLooper_;
@@ -984,6 +1006,30 @@ class QuicTransportBase : public QuicSocket, QuicStreamPrioritiesObserver {
 
   bool isTimeoutScheduled(QuicTimerCallback* callback) const;
 
+  /**
+   * Helper function to validate that the number of ECN packet marks match the
+   * expected value, depending on the ECN state of the connection.
+   *
+   * If ECN is enabled, this function validates it's working correctly. If ECN
+   * is not enabled or has already failed validation, this function does
+   * nothing.
+   */
+  void validateECNState();
+
+  WriteQuicDataResult handleInitialWriteDataCommon(
+      const ConnectionId& srcConnId,
+      const ConnectionId& dstConnId,
+      uint64_t packetLimit,
+      const std::string& token = "");
+
+  WriteQuicDataResult handleHandshakeWriteDataCommon(
+      const ConnectionId& srcConnId,
+      const ConnectionId& dstConnId,
+      uint64_t packetLimit);
+
+  void onSocketWritable() noexcept override;
+  void maybeStopWriteLooperAndArmSocketWritableEvent();
+
  private:
   /**
    * Helper functions to handle new streams.
