@@ -126,7 +126,7 @@ WriteQuicDataResult writeQuicDataToSocketImpl(
     uint64_t packetLimit,
     bool exceptCryptoStream,
     TimePoint writeLoopBeginTime) {
-  auto builder = ShortHeaderBuilder();
+  auto builder = ShortHeaderBuilder(connection.oneRttWritePhase);
   WriteQuicDataResult result;
   auto& packetsWritten = result.packetsWritten;
   auto& probesWritten = result.probesWritten;
@@ -261,7 +261,7 @@ DataPathResult continuousMemoryBuildScheduleEncrypt(
     }
     return DataPathResult::makeBuildFailure();
   }
-  if (!packet->body) {
+  if (packet->body.empty()) {
     // No more space remaining.
     rollbackBuf();
     ioBufBatch.flush();
@@ -270,21 +270,20 @@ DataPathResult continuousMemoryBuildScheduleEncrypt(
     }
     return DataPathResult::makeBuildFailure();
   }
-  CHECK(!packet->header->isChained());
-  auto headerLen = packet->header->length();
+  CHECK(!packet->header.isChained());
+  auto headerLen = packet->header.length();
   buf = connection.bufAccessor->obtain();
   CHECK(
-      packet->body->data() > buf->data() &&
-      packet->body->tail() <= buf->tail());
+      packet->body.data() > buf->data() && packet->body.tail() <= buf->tail());
   CHECK(
-      packet->header->data() >= buf->data() &&
-      packet->header->tail() < buf->tail());
+      packet->header.data() >= buf->data() &&
+      packet->header.tail() < buf->tail());
   // Trim off everything before the current packet, and the header length, so
   // buf's data starts from the body part of buf.
   buf->trimStart(prevSize + headerLen);
   // buf and packetBuf is actually the same.
   auto packetBuf =
-      aead.inplaceEncrypt(std::move(buf), packet->header.get(), packetNum);
+      aead.inplaceEncrypt(std::move(buf), &packet->header, packetNum);
   CHECK(packetBuf->headroom() == headerLen + prevSize);
   // Include header back.
   packetBuf->prepend(headerLen);
@@ -309,11 +308,6 @@ DataPathResult continuousMemoryBuildScheduleEncrypt(
   }
   // TODO: I think we should add an API that doesn't need a buffer.
   bool ret = ioBufBatch.write(nullptr /* no need to pass buf */, encodedSize);
-  // update stats and connection
-  if (ret) {
-    QUIC_STATS(connection.statsCallback, onWrite, encodedSize);
-    QUIC_STATS(connection.statsCallback, onPacketSent);
-  }
   return DataPathResult::makeWriteResult(
       ret, std::move(result), encodedSize, encodedBodySize);
 }
@@ -345,7 +339,7 @@ DataPathResult iobufChainBasedBuildScheduleEncrypt(
     }
     return DataPathResult::makeBuildFailure();
   }
-  if (!packet->body) {
+  if (packet->body.empty()) {
     // No more space remaining.
     ioBufBatch.flush();
     if (connection.loopDetectorCallback) {
@@ -353,20 +347,20 @@ DataPathResult iobufChainBasedBuildScheduleEncrypt(
     }
     return DataPathResult::makeBuildFailure();
   }
-  packet->header->coalesce();
-  auto headerLen = packet->header->length();
-  auto bodyLen = packet->body->computeChainDataLength();
+  packet->header.coalesce();
+  auto headerLen = packet->header.length();
+  auto bodyLen = packet->body.computeChainDataLength();
   auto unencrypted = folly::IOBuf::createCombined(
       headerLen + bodyLen + aead.getCipherOverhead());
-  auto bodyCursor = folly::io::Cursor(packet->body.get());
+  auto bodyCursor = folly::io::Cursor(&packet->body);
   bodyCursor.pull(unencrypted->writableData() + headerLen, bodyLen);
   unencrypted->advance(headerLen);
   unencrypted->append(bodyLen);
-  auto packetBuf = aead.inplaceEncrypt(
-      std::move(unencrypted), packet->header.get(), packetNum);
+  auto packetBuf =
+      aead.inplaceEncrypt(std::move(unencrypted), &packet->header, packetNum);
   DCHECK(packetBuf->headroom() == headerLen);
   packetBuf->clear();
-  auto headerCursor = folly::io::Cursor(packet->header.get());
+  auto headerCursor = folly::io::Cursor(&packet->header);
   headerCursor.pull(packetBuf->writableData(), headerLen);
   packetBuf->append(headerLen + bodyLen + aead.getCipherOverhead());
 
@@ -385,11 +379,6 @@ DataPathResult iobufChainBasedBuildScheduleEncrypt(
             << " encodedBodySize=" << encodedBodySize;
   }
   bool ret = ioBufBatch.write(std::move(packetBuf), encodedSize);
-  if (ret) {
-    // update stats and connection
-    QUIC_STATS(connection.statsCallback, onWrite, encodedSize);
-    QUIC_STATS(connection.statsCallback, onPacketSent);
-  }
   return DataPathResult::makeWriteResult(
       ret, std::move(result), encodedSize, encodedBodySize);
 }
@@ -468,7 +457,7 @@ void handleRetransmissionWritten(
     uint64_t frameOffset,
     uint64_t frameLen,
     bool frameFin,
-    std::deque<StreamBuffer>::iterator lossBufferIter) {
+    CircularDeque<StreamBuffer>::iterator lossBufferIter) {
   auto bufferLen = lossBufferIter->data.chainLength();
   Buf bufWritten;
   if (frameLen == bufferLen && frameFin == lossBufferIter->eof) {
@@ -845,6 +834,20 @@ void updateConnection(
   conn.lossState.totalStreamBytesSent += streamBytesSent;
   conn.lossState.totalNewStreamBytesSent += newStreamBytesSent;
 
+  // Count the number of packets sent in the current phase.
+  // This is used to initiate key updates if enabled.
+  if (packet.header.getProtectionType() == conn.oneRttWritePhase) {
+    if (conn.oneRttWritePendingVerification &&
+        conn.oneRttWritePacketsSentInCurrentPhase == 0) {
+      // This is the first packet in the new phase after we have initiated a key
+      // update. We need to keep track of it to confirm the peer acks it in the
+      // same phase.
+      conn.oneRttWritePendingVerificationPacketNumber =
+          packet.header.getPacketSequenceNum();
+    }
+    conn.oneRttWritePacketsSentInCurrentPhase++;
+  }
+
   if (!retransmittable && !isPing) {
     DCHECK(!packetEvent);
     return;
@@ -879,14 +882,12 @@ void updateConnection(
       // conn.lossState.inflightBytes isn't updated until below
       // conn.outstandings.numOutstanding() + 1 since we're emplacing here
       conn.lossState.totalBytesSent,
-      conn.lossState.totalBodyBytesSent,
       conn.lossState.inflightBytes + encodedSize,
-      conn.outstandings.numOutstanding() + 1,
       conn.lossState,
       conn.writeCount,
       std::move(detailsPerStream),
       conn.appLimitedTracker.getTotalAppLimitedTime(),
-      packetDestroyFn);
+      std::move(packetDestroyFn));
 
   pkt.metadata.cmsgs = conn.socketCmsgsState.additionalCmsgs;
 
@@ -1010,13 +1011,14 @@ HeaderBuilder LongHeaderBuilder(LongHeader::Types packetType) {
   };
 }
 
-HeaderBuilder ShortHeaderBuilder() {
-  return [](const ConnectionId& /* srcConnId */,
-            const ConnectionId& dstConnId,
-            PacketNum packetNum,
-            QuicVersion,
-            const std::string&) {
-    return ShortHeader(ProtectionType::KeyPhaseZero, dstConnId, packetNum);
+HeaderBuilder ShortHeaderBuilder(ProtectionType keyPhase) {
+  return [keyPhase](
+             const ConnectionId& /* srcConnId */,
+             const ConnectionId& dstConnId,
+             PacketNum packetNum,
+             QuicVersion,
+             const std::string&) {
+    return ShortHeader(keyPhase, dstConnId, packetNum);
   };
 }
 
@@ -1262,22 +1264,23 @@ void writeCloseCommon(
     return;
   }
   auto packet = std::move(packetBuilder).buildPacket();
-  packet.header->coalesce();
-  packet.body->reserve(0, aead.getCipherOverhead());
-  CHECK_GE(packet.body->tailroom(), aead.getCipherOverhead());
-  auto body = aead.inplaceEncrypt(
-      std::move(packet.body), packet.header.get(), packetNum);
-  body->coalesce();
+  packet.header.coalesce();
+  packet.body.reserve(0, aead.getCipherOverhead());
+  CHECK_GE(packet.body.tailroom(), aead.getCipherOverhead());
+  auto bufUniquePtr = packet.body.clone();
+  bufUniquePtr =
+      aead.inplaceEncrypt(std::move(bufUniquePtr), &packet.header, packetNum);
+  bufUniquePtr->coalesce();
   encryptPacketHeader(
       headerForm,
-      packet.header->writableData(),
-      packet.header->length(),
-      body->data(),
-      body->length(),
+      packet.header.writableData(),
+      packet.header.length(),
+      bufUniquePtr->data(),
+      bufUniquePtr->length(),
       headerCipher);
-  auto packetBuf = std::move(packet.header);
-  packetBuf->prependChain(std::move(body));
-  auto packetSize = packetBuf->computeChainDataLength();
+  folly::IOBuf packetBuf(std::move(packet.header));
+  packetBuf.prependChain(std::move(bufUniquePtr));
+  auto packetSize = packetBuf.computeChainDataLength();
   if (connection.qLogger) {
     connection.qLogger->addPacket(packet.packet, packetSize);
   }
@@ -1287,7 +1290,9 @@ void writeCloseCommon(
   // Increment the sequence number.
   increaseNextPacketNum(connection, pnSpace);
   // best effort writing to the socket, ignore any errors.
-  auto ret = sock.write(connection.peerAddress, packetBuf);
+
+  Buf packetBufPtr = packetBuf.clone();
+  auto ret = sock.write(connection.peerAddress, packetBufPtr);
   connection.lossState.totalBytesSent += packetSize;
   if (ret < 0) {
     VLOG(4) << "Error writing connection close " << folly::errnoStr(errno)
@@ -1336,7 +1341,7 @@ void writeShortClose(
     const Aead& aead,
     const PacketNumberCipher& headerCipher) {
   auto header = ShortHeader(
-      ProtectionType::KeyPhaseZero,
+      connection.oneRttWritePhase,
       connId,
       getNextPacketNum(connection, PacketNumberSpace::AppData));
   writeCloseCommon(
@@ -1455,7 +1460,12 @@ WriteQuicDataResult writeConnectionDataToSocket(
     connection.writeDebugState.noWriteReason = NoWriteReason::WRITE_OK;
   }
 
-  if (!scheduler.hasData()) {
+  // Note: if a write is pending, it will be taken over by the batch writer when
+  // it's created. So this check has to be done before creating the batch
+  // writer.
+  bool pendingBufferedWrite = hasBufferedDataToWrite(connection);
+
+  if (!scheduler.hasData() && !pendingBufferedWrite) {
     if (connection.loopDetectorCallback) {
       connection.writeDebugState.noWriteReason = NoWriteReason::EMPTY_SCHEDULER;
     }
@@ -1487,6 +1497,7 @@ WriteQuicDataResult writeConnectionDataToSocket(
   auto batchWriter = BatchWriterFactory::makeBatchWriter(
       connection.transportSettings.batchingMode,
       connection.transportSettings.maxBatchSize,
+      connection.transportSettings.enableWriterBackpressure,
       connection.transportSettings.dataPathType,
       connection,
       *connection.gsoSupported);
@@ -1501,12 +1512,40 @@ WriteQuicDataResult writeConnectionDataToSocket(
       connection.statsCallback,
       happyEyeballsState);
 
+  // If we have a pending write to retry. Flush that first and make sure it
+  // succeeds before scheduling any new data.
+  if (pendingBufferedWrite) {
+    if (!ioBufBatch.flush()) {
+      // Could not flush retried data. Return empty write result and wait for
+      // next retry.
+      return {0, 0, 0};
+    }
+  }
+
   auto batchSize = connection.transportSettings.batchingMode ==
           QuicBatchingMode::BATCHING_MODE_NONE
       ? connection.transportSettings.writeConnectionDataPacketsLimit
       : connection.transportSettings.maxBatchSize;
 
   uint64_t bytesWritten = 0;
+  uint64_t shortHeaderPadding = 0;
+  uint64_t shortHeaderPaddingCount = 0;
+  SCOPE_EXIT {
+    auto nSent = ioBufBatch.getPktSent();
+    if (nSent > 0) {
+      QUIC_STATS(connection.statsCallback, onPacketsSent, nSent);
+      QUIC_STATS(connection.statsCallback, onWrite, bytesWritten);
+      if (shortHeaderPadding > 0) {
+        QUIC_STATS(
+            connection.statsCallback,
+            onShortHeaderPaddingBatch,
+            shortHeaderPaddingCount,
+            shortHeaderPadding);
+      }
+    }
+  };
+
+  quic::TimePoint sentTime = Clock::now();
 
   while (scheduler.hasData() && ioBufBatch.getPktSent() < packetLimit &&
          ((ioBufBatch.getPktSent() < batchSize) ||
@@ -1550,13 +1589,17 @@ WriteQuicDataResult writeConnectionDataToSocket(
     // pretend write was also successful but packet is lost somewhere in the
     // network.
     bytesWritten += ret.encodedSize;
+    if (ret.result && ret.result->shortHeaderPadding > 0) {
+      shortHeaderPaddingCount++;
+      shortHeaderPadding += ret.result->shortHeaderPadding;
+    }
 
     auto& result = ret.result;
     updateConnection(
         connection,
         std::move(result->packetEvent),
         std::move(result->packet->packet),
-        Clock::now(),
+        sentTime,
         folly::to<uint32_t>(ret.encodedSize),
         folly::to<uint32_t>(ret.encodedBodySize),
         false /* isDSRPacket */);
@@ -1707,6 +1750,11 @@ WriteDataReason shouldWriteData(/*const*/ QuicConnectionStateBase& conn) {
     QUIC_STATS(conn.statsCallback, onCwndBlocked);
     return WriteDataReason::NO_WRITE;
   }
+
+  if (hasBufferedDataToWrite(conn)) {
+    return WriteDataReason::BUFFERED_WRITE;
+  }
+
   return hasNonAckDataToWrite(conn);
 }
 
@@ -1729,11 +1777,14 @@ bool hasAckDataToWrite(const QuicConnectionStateBase& conn) {
   return writeAcks;
 }
 
+bool hasBufferedDataToWrite(const QuicConnectionStateBase& conn) {
+  return (bool)conn.pendingWriteBatch_.buf;
+}
+
 WriteDataReason hasNonAckDataToWrite(const QuicConnectionStateBase& conn) {
   if (cryptoHasWritableData(conn)) {
     VLOG(10) << nodeToString(conn.nodeType)
-             << " needs write because of crypto stream"
-             << " " << conn;
+             << " needs write because of crypto stream" << " " << conn;
     return WriteDataReason::CRYPTO_STREAM;
   }
   if (!conn.oneRttWriteCipher &&
@@ -1818,7 +1869,10 @@ void implicitAckCryptoStream(
       conn,
       packetNumSpace,
       implicitAck,
-      [&](auto&, auto& packetFrame, auto&) {
+      [](const auto&) {
+        // ackedPacketVisitor. No action needed.
+      },
+      [&](auto&, auto& packetFrame) {
         switch (packetFrame.type()) {
           case QuicWriteFrame::Type::WriteCryptoFrame: {
             const WriteCryptoFrame& frame = *packetFrame.asWriteCryptoFrame();
@@ -1901,4 +1955,86 @@ bool toWriteAppDataAcks(const quic::QuicConnectionStateBase& conn) {
       hasAcksToSchedule(conn.ackStates.appDataAckState) &&
       conn.ackStates.appDataAckState.needsToSendAckImmediately);
 }
+
+void updateOneRttWriteCipher(
+    quic::QuicConnectionStateBase& conn,
+    std::unique_ptr<Aead> aead,
+    ProtectionType oneRttPhase) {
+  CHECK(
+      oneRttPhase == ProtectionType::KeyPhaseZero ||
+      oneRttPhase == ProtectionType::KeyPhaseOne);
+  CHECK(oneRttPhase != conn.oneRttWritePhase)
+      << "Cannot replace cipher for current write phase";
+  conn.oneRttWriteCipher = std::move(aead);
+  conn.oneRttWritePhase = oneRttPhase;
+  conn.oneRttWritePacketsSentInCurrentPhase = 0;
+}
+
+void maybeHandleIncomingKeyUpdate(QuicConnectionStateBase& conn) {
+  if (conn.readCodec->getCurrentOneRttReadPhase() != conn.oneRttWritePhase) {
+    // Peer has initiated a key update.
+    updateOneRttWriteCipher(
+        conn,
+        conn.handshakeLayer->getNextOneRttWriteCipher(),
+        conn.readCodec->getCurrentOneRttReadPhase());
+
+    conn.readCodec->setNextOneRttReadCipher(
+        conn.handshakeLayer->getNextOneRttReadCipher());
+  }
+}
+
+void maybeInitiateKeyUpdate(QuicConnectionStateBase& conn) {
+  if (conn.transportSettings.initiateKeyUpdate) {
+    auto packetsBeforeNextUpdate =
+        conn.transportSettings.firstKeyUpdatePacketCount
+        ? conn.transportSettings.firstKeyUpdatePacketCount.value()
+        : conn.transportSettings.keyUpdatePacketCountInterval;
+
+    if ((conn.oneRttWritePacketsSentInCurrentPhase > packetsBeforeNextUpdate) &&
+        conn.readCodec->canInitiateKeyUpdate()) {
+      QUIC_STATS(conn.statsCallback, onKeyUpdateAttemptInitiated);
+      conn.readCodec->advanceOneRttReadPhase();
+      conn.transportSettings.firstKeyUpdatePacketCount.clear();
+
+      updateOneRttWriteCipher(
+          conn,
+          conn.handshakeLayer->getNextOneRttWriteCipher(),
+          conn.readCodec->getCurrentOneRttReadPhase());
+      conn.readCodec->setNextOneRttReadCipher(
+          conn.handshakeLayer->getNextOneRttReadCipher());
+      // Signal the transport that a key update has been initiated.
+      conn.oneRttWritePendingVerification = true;
+      conn.oneRttWritePendingVerificationPacketNumber.clear();
+    }
+  }
+}
+
+void maybeVerifyPendingKeyUpdate(
+    QuicConnectionStateBase& conn,
+    const OutstandingPacketWrapper& outstandingPacket,
+    const RegularQuicPacket& ackPacket) {
+  if (!(protectionTypeToEncryptionLevel(
+            outstandingPacket.packet.header.getProtectionType()) ==
+        EncryptionLevel::AppData)) {
+    // This is not an app data packet. We can't have initiated a key update yet.
+    return;
+  }
+
+  if (conn.oneRttWritePendingVerificationPacketNumber &&
+      outstandingPacket.packet.header.getPacketSequenceNum() >=
+          conn.oneRttWritePendingVerificationPacketNumber.value()) {
+    // There is a pending key update. This packet should be acked in
+    // the current phase.
+    if (ackPacket.header.getProtectionType() == conn.oneRttWritePhase) {
+      // Key update is verified.
+      conn.oneRttWritePendingVerificationPacketNumber.clear();
+      conn.oneRttWritePendingVerification = false;
+    } else {
+      throw QuicTransportException(
+          "Packet with key update was acked in the wrong phase",
+          TransportErrorCode::CRYPTO_ERROR);
+    }
+  }
+}
+
 } // namespace quic
