@@ -66,8 +66,11 @@ class QuicTypedTransportTestBase : protected QuicTransportTestClass {
     //
     // we include "lost" to protect against the unusual case of the test somehow
     // causing a packet that was just written to be immediately marked lost.
-    const auto it = quic::getLastOutstandingPacketIncludingLost(
-        getNonConstConn(), PacketNumberSpace::AppData);
+    const auto it = quic::getLastOutstandingPacket(
+        getNonConstConn(),
+        PacketNumberSpace::AppData,
+        true /* includeDeclaredLost */);
+
     if (it == getConn().outstandings.packets.rend()) {
       return folly::none;
     }
@@ -157,7 +160,8 @@ class QuicTypedTransportTestBase : protected QuicTransportTestClass {
                 pnSpace,
                 oldestOutstandingPkt->packet.header.getPacketSequenceNum(),
                 newestOutstandingPkt->packet.header.getPacketSequenceNum()),
-            recvTime),
+            recvTime,
+            0),
         false /* loopForWrites */);
   }
 
@@ -177,9 +181,10 @@ class QuicTypedTransportTestBase : protected QuicTransportTestClass {
   void deliverPacket(
       Buf&& buf,
       quic::TimePoint recvTime = TimePoint::clock::now(),
+      uint8_t tosValue = 0,
       bool loopForWrites = true) {
     QuicTransportTestClass::deliverData(
-        NetworkData(std::move(buf), recvTime), loopForWrites);
+        NetworkData(std::move(buf), recvTime, tosValue), loopForWrites);
   }
 
   /**
@@ -187,8 +192,10 @@ class QuicTypedTransportTestBase : protected QuicTransportTestClass {
    */
   void deliverPacketNoWrites(
       Buf&& buf,
-      quic::TimePoint recvTime = TimePoint::clock::now()) {
-    deliverPacket(std::move(buf), recvTime, false /* loopForWrites */);
+      quic::TimePoint recvTime = TimePoint::clock::now(),
+      uint8_t tosValue = 0) {
+    deliverPacket(
+        std::move(buf), recvTime, tosValue, false /* loopForWrites */);
   }
 
   /**
@@ -197,9 +204,20 @@ class QuicTypedTransportTestBase : protected QuicTransportTestClass {
   void deliverPackets(
       std::vector<Buf>&& bufs,
       quic::TimePoint recvTime = TimePoint::clock::now(),
+      uint8_t tosValue = 0,
       bool loopForWrites = true) {
-    QuicTransportTestClass::deliverData(
-        NetworkData(std::move(bufs), recvTime), loopForWrites);
+    auto networkData = NetworkData();
+    // This overrides the timing in added packets.
+    networkData.setReceiveTimePoint(recvTime);
+    for (auto& buf : bufs) {
+      auto udpPacket = ReceivedUdpPacket(
+          std::move(buf),
+          ReceivedUdpPacket::Timings{}, // NetworkData receiveTimePoint will
+                                        // override this
+          tosValue);
+      networkData.addPacket(std::move(udpPacket));
+    }
+    QuicTransportTestClass::deliverData(std::move(networkData), loopForWrites);
   }
 
   /**
@@ -216,7 +234,9 @@ class QuicTypedTransportTestBase : protected QuicTransportTestClass {
    */
   quic::Buf buildPeerPacketWithStreamData(
       const quic::StreamId streamId,
-      Buf data) {
+      Buf data,
+      folly::Optional<ProtectionType> shortHeaderProtectionOverride =
+          folly::none) {
     auto buf = quic::test::packetToBuf(createStreamPacket(
         getSrcConnectionId(),
         getDstConnectionId(),
@@ -229,7 +249,8 @@ class QuicTypedTransportTestBase : protected QuicTransportTestClass {
         // // should meet the needs of the majority of tests...
         // getConn().ackStates.appDataAckState.largestAckedByPeer.value_or(0),
         folly::none /* longHeaderOverride */,
-        false /* eof */));
+        false /* eof */,
+        shortHeaderProtectionOverride));
     buf->coalesce();
     return buf;
   }
