@@ -123,8 +123,7 @@ void QuicClientTransport::processUdpPacket(
     ReceivedUdpPacket&& udpPacket) {
   // Process the arriving UDP packet, which may have coalesced QUIC packets.
   {
-    BufQueue udpData;
-    udpData.append(std::move(udpPacket.buf));
+    BufQueue& udpData = udpPacket.buf;
 
     if (!conn_->version) {
       // We only check for version negotiation packets before the version
@@ -138,14 +137,14 @@ void QuicClientTransport::processUdpPacket(
 
         throw QuicInternalException(
             "Received version negotiation packet",
-            LocalErrorCode::CONNECTION_ABANDONED);
+            LocalErrorCode::NEW_VERSION_NEGOTIATED);
       }
     }
 
     for (uint16_t processedPackets = 0;
          !udpData.empty() && processedPackets < kMaxNumCoalescedPackets;
          processedPackets++) {
-      processUdpPacketData(peer, udpPacket.timings, udpData);
+      processUdpPacketData(peer, udpPacket);
     }
     VLOG_IF(4, !udpData.empty())
         << "Leaving " << udpData.chainLength()
@@ -156,23 +155,23 @@ void QuicClientTransport::processUdpPacket(
   // Process any deferred pending 1RTT and handshake packets if we have keys.
   if (conn_->readCodec->getOneRttReadCipher() &&
       !clientConn_->pendingOneRttData.empty()) {
-    BufQueue pendingPacket;
-    for (auto& pendingData : clientConn_->pendingOneRttData) {
-      pendingPacket.append(std::move(pendingData.udpPacket.buf));
-      processUdpPacketData(
-          pendingData.peer, pendingData.udpPacket.timings, pendingPacket);
-      pendingPacket.move();
+    for (auto& pendingPacket : clientConn_->pendingOneRttData) {
+      // The first loop should try to process any leftover data in the incoming
+      // buffer.
+      pendingPacket.udpPacket.buf.append(udpPacket.buf.move());
+
+      processUdpPacketData(pendingPacket.peer, pendingPacket.udpPacket);
     }
     clientConn_->pendingOneRttData.clear();
   }
   if (conn_->readCodec->getHandshakeReadCipher() &&
       !clientConn_->pendingHandshakeData.empty()) {
-    BufQueue pendingPacket;
-    for (auto& pendingData : clientConn_->pendingHandshakeData) {
-      pendingPacket.append(std::move(pendingData.udpPacket.buf));
-      processUdpPacketData(
-          pendingData.peer, pendingData.udpPacket.timings, pendingPacket);
-      pendingPacket.move();
+    for (auto& pendingPacket : clientConn_->pendingHandshakeData) {
+      // The first loop should try to process any leftover data in the incoming
+      // buffer.
+      pendingPacket.udpPacket.buf.append(udpPacket.buf.move());
+
+      processUdpPacketData(pendingPacket.peer, pendingPacket.udpPacket);
     }
     clientConn_->pendingHandshakeData.clear();
   }
@@ -180,14 +179,13 @@ void QuicClientTransport::processUdpPacket(
 
 void QuicClientTransport::processUdpPacketData(
     const folly::SocketAddress& peer,
-    const ReceivedUdpPacket::Timings& udpPacketTimings,
-    BufQueue& udpPacketData) {
-  auto packetSize = udpPacketData.chainLength();
+    ReceivedUdpPacket& udpPacket) {
+  auto packetSize = udpPacket.buf.chainLength();
   if (packetSize == 0) {
     return;
   }
   auto parsedPacket = conn_->readCodec->parsePacket(
-      udpPacketData, conn_->ackStates, conn_->clientConnectionId->size());
+      udpPacket.buf, conn_->ackStates, conn_->clientConnectionId->size());
   StatelessReset* statelessReset = parsedPacket.statelessReset();
   if (statelessReset) {
     const auto& token = clientConn_->statelessResetToken;
@@ -270,7 +268,9 @@ void QuicClientTransport::processUdpPacketData(
         : clientConn_->pendingHandshakeData;
     pendingData.emplace_back(
         ReceivedUdpPacket(
-            std::move(cipherUnavailable->packet), udpPacketTimings),
+            std::move(cipherUnavailable->packet),
+            udpPacket.timings,
+            udpPacket.tosValue),
         peer);
     if (conn_->qLogger) {
       conn_->qLogger->addPacketBuffered(
@@ -375,7 +375,7 @@ void QuicClientTransport::processUdpPacketData(
   // Add the packet to the AckState associated with the packet number space.
   auto& ackState = getAckState(*conn_, pnSpace);
   uint64_t distanceFromExpectedPacketNum =
-      addPacketToAckState(*conn_, ackState, packetNum, udpPacketTimings);
+      addPacketToAckState(*conn_, ackState, packetNum, udpPacket);
   if (distanceFromExpectedPacketNum > 0) {
     QUIC_STATS(conn_->statsCallback, onOutOfOrderPacketReceived);
   }
@@ -383,6 +383,78 @@ void QuicClientTransport::processUdpPacketData(
   bool pktHasRetransmittableData = false;
   bool pktHasCryptoData = false;
 
+  AckedPacketVisitor ackedPacketVisitor =
+      [&](const OutstandingPacketWrapper& outstandingPacket) {
+        auto outstandingProtectionType =
+            outstandingPacket.packet.header.getProtectionType();
+        if (outstandingProtectionType == ProtectionType::KeyPhaseZero) {
+          // If we received an ack for data that we sent in 1-rtt from
+          // the server, we can assume that the server had successfully
+          // derived the 1-rtt keys and hence received the client
+          // finished message. We can mark the handshake as confirmed and
+          // drop the handshake cipher and outstanding packets after the
+          // processing loop.
+          conn_->handshakeLayer->handshakeConfirmed();
+        }
+        maybeVerifyPendingKeyUpdate(*conn_, outstandingPacket, regularPacket);
+      };
+
+  AckedFrameVisitor ackedFrameVisitor =
+      [&](const OutstandingPacketWrapper& outstandingPacket,
+          const QuicWriteFrame& packetFrame) {
+        auto outstandingProtectionType =
+            outstandingPacket.packet.header.getProtectionType();
+        switch (packetFrame.type()) {
+          case QuicWriteFrame::Type::WriteAckFrame: {
+            const WriteAckFrame& frame = *packetFrame.asWriteAckFrame();
+            DCHECK(!frame.ackBlocks.empty());
+            VLOG(4) << "Client received ack for largestAcked="
+                    << frame.ackBlocks.front().end << " " << *this;
+            commonAckVisitorForAckFrame(ackState, frame);
+            break;
+          }
+          case QuicWriteFrame::Type::RstStreamFrame: {
+            const RstStreamFrame& frame = *packetFrame.asRstStreamFrame();
+            VLOG(4) << "Client received ack for reset frame stream="
+                    << frame.streamId << " " << *this;
+
+            auto stream = conn_->streamManager->getStream(frame.streamId);
+            if (stream) {
+              sendRstAckSMHandler(*stream);
+            }
+            break;
+          }
+          case QuicWriteFrame::Type::WriteStreamFrame: {
+            const WriteStreamFrame& frame = *packetFrame.asWriteStreamFrame();
+
+            auto ackedStream = conn_->streamManager->getStream(frame.streamId);
+            VLOG(4) << "Client got ack for stream=" << frame.streamId
+                    << " offset=" << frame.offset << " fin=" << frame.fin
+                    << " data=" << frame.len
+                    << " closed=" << (ackedStream == nullptr) << " " << *this;
+            if (ackedStream) {
+              sendAckSMHandler(*ackedStream, frame);
+            }
+            break;
+          }
+          case QuicWriteFrame::Type::WriteCryptoFrame: {
+            const WriteCryptoFrame& frame = *packetFrame.asWriteCryptoFrame();
+            auto cryptoStream = getCryptoStream(
+                *conn_->cryptoState,
+                protectionTypeToEncryptionLevel(outstandingProtectionType));
+            processCryptoStreamAck(*cryptoStream, frame.offset, frame.len);
+            break;
+          }
+          case QuicWriteFrame::Type::PingFrame:
+            conn_->pendingEvents.cancelPingTimeout = true;
+            break;
+          case QuicWriteFrame::Type::QuicSimpleFrame:
+          default:
+            // ignore other frames.
+            break;
+        }
+      };
+
   for (auto& quicFrame : regularPacket.frames) {
     switch (quicFrame.type()) {
       case QuicFrame::Type::ReadAckFrame: {
@@ -393,78 +465,10 @@ void QuicClientTransport::processUdpPacketData(
             *conn_,
             pnSpace,
             ackFrame,
-            [&](const OutstandingPacketWrapper& outstandingPacket,
-                const QuicWriteFrame& packetFrame,
-                const ReadAckFrame&) {
-              auto outstandingProtectionType =
-                  outstandingPacket.packet.header.getProtectionType();
-              if (outstandingProtectionType == ProtectionType::KeyPhaseZero) {
-                // If we received an ack for data that we sent in 1-rtt from
-                // the server, we can assume that the server had successfully
-                // derived the 1-rtt keys and hence received the client
-                // finished message. We can mark the handshake as confirmed and
-                // drop the handshake cipher and outstanding packets after the
-                // processing loop.
-                conn_->handshakeLayer->handshakeConfirmed();
-              }
-              switch (packetFrame.type()) {
-                case QuicWriteFrame::Type::WriteAckFrame: {
-                  const WriteAckFrame& frame = *packetFrame.asWriteAckFrame();
-                  DCHECK(!frame.ackBlocks.empty());
-                  VLOG(4) << "Client received ack for largestAcked="
-                          << frame.ackBlocks.front().end << " " << *this;
-                  commonAckVisitorForAckFrame(ackState, frame);
-                  break;
-                }
-                case QuicWriteFrame::Type::RstStreamFrame: {
-                  const RstStreamFrame& frame = *packetFrame.asRstStreamFrame();
-                  VLOG(4) << "Client received ack for reset frame stream="
-                          << frame.streamId << " " << *this;
-
-                  auto stream = conn_->streamManager->getStream(frame.streamId);
-                  if (stream) {
-                    sendRstAckSMHandler(*stream);
-                  }
-                  break;
-                }
-                case QuicWriteFrame::Type::WriteStreamFrame: {
-                  const WriteStreamFrame& frame =
-                      *packetFrame.asWriteStreamFrame();
-
-                  auto ackedStream =
-                      conn_->streamManager->getStream(frame.streamId);
-                  VLOG(4) << "Client got ack for stream=" << frame.streamId
-                          << " offset=" << frame.offset << " fin=" << frame.fin
-                          << " data=" << frame.len
-                          << " closed=" << (ackedStream == nullptr) << " "
-                          << *this;
-                  if (ackedStream) {
-                    sendAckSMHandler(*ackedStream, frame);
-                  }
-                  break;
-                }
-                case QuicWriteFrame::Type::WriteCryptoFrame: {
-                  const WriteCryptoFrame& frame =
-                      *packetFrame.asWriteCryptoFrame();
-                  auto cryptoStream = getCryptoStream(
-                      *conn_->cryptoState,
-                      protectionTypeToEncryptionLevel(
-                          outstandingProtectionType));
-                  processCryptoStreamAck(
-                      *cryptoStream, frame.offset, frame.len);
-                  break;
-                }
-                case QuicWriteFrame::Type::PingFrame:
-                  conn_->pendingEvents.cancelPingTimeout = true;
-                  break;
-                case QuicWriteFrame::Type::QuicSimpleFrame:
-                default:
-                  // ignore other frames.
-                  break;
-              }
-            },
+            ackedPacketVisitor,
+            ackedFrameVisitor,
             markPacketLoss,
-            udpPacketTimings.receiveTimePoint));
+            udpPacket.timings.receiveTimePoint));
         break;
       }
       case QuicFrame::Type::RstStreamFrame: {
@@ -614,12 +618,12 @@ void QuicClientTransport::processUdpPacketData(
       }
       case QuicFrame::Type::DatagramFrame: {
         DatagramFrame& frame = *quicFrame.asDatagramFrame();
-        VLOG(10) << "Client received datagram data: "
-                 << "len=" << frame.length << " " << *this;
+        VLOG(10) << "Client received datagram data: " << "len=" << frame.length
+                 << " " << *this;
         // Datagram isn't retransmittable. But we would like to ack them early.
         // So, make Datagram frames count towards ack policy
         pktHasRetransmittableData = true;
-        handleDatagram(*conn_, frame, udpPacketTimings.receiveTimePoint);
+        handleDatagram(*conn_, frame, udpPacket.timings.receiveTimePoint);
         break;
       }
       case QuicFrame::Type::ImmediateAckFrame: {
@@ -652,6 +656,8 @@ void QuicClientTransport::processUdpPacketData(
     handshakeConfirmed(*conn_);
   }
 
+  maybeHandleIncomingKeyUpdate(*conn_);
+
   // Try reading bytes off of crypto, and performing a handshake.
   auto cryptoData = readDataFromCryptoStream(
       *getCryptoStream(*conn_->cryptoState, encryptionLevel));
@@ -676,6 +682,11 @@ void QuicClientTransport::processUdpPacketData(
         }
         QUIC_STATS(conn_->statsCallback, onZeroRttRejected);
         handshakeLayer->removePsk(hostname_);
+        if (!handshakeLayer->getCanResendZeroRtt().value_or(false)) {
+          throw QuicTransportException(
+              "Zero-rtt attempted but the early parameters do not match the handshake parameters",
+              TransportErrorCode::TRANSPORT_PARAMETER_ERROR);
+        }
       } else if (clientConn_->zeroRttRejected.has_value()) {
         if (conn_->qLogger) {
           conn_->qLogger->addTransportStateUpdate(kZeroRttAccepted);
@@ -778,6 +789,9 @@ void QuicClientTransport::processUdpPacketData(
       if (clientConn_->statelessResetToken) {
         conn_->readCodec->setStatelessResetToken(
             clientConn_->statelessResetToken.value());
+        auto& cryptoFactory = handshakeLayer->getCryptoFactory();
+        conn_->readCodec->setCryptoEqual(
+            cryptoFactory.getCryptoEqualFunction());
       }
     }
 
@@ -849,9 +863,8 @@ void QuicClientTransport::onReadData(
 void QuicClientTransport::writeData() {
   QuicVersion version = conn_->version.value_or(*conn_->originalVersion);
   const ConnectionId& srcConnId = *conn_->clientConnectionId;
-  const ConnectionId& destConnId = conn_->serverConnectionId
-      ? *conn_->serverConnectionId
-      : *clientConn_->initialDestinationConnectionId;
+  const ConnectionId& destConnId = conn_->serverConnectionId.value_or(
+      *clientConn_->initialDestinationConnectionId);
 
   if (closeState_ == CloseState::CLOSED) {
     auto rtt = clientConn_->lossState.srtt == 0us
@@ -911,61 +924,23 @@ void QuicClientTransport::writeData() {
   // use.
   SCOPE_EXIT {
     conn_->pendingEvents.numProbePackets = {};
+    maybeInitiateKeyUpdate(*conn_);
   };
   if (conn_->initialWriteCipher) {
-    auto& initialCryptoStream =
-        *getCryptoStream(*conn_->cryptoState, EncryptionLevel::Initial);
-    CryptoStreamScheduler initialScheduler(*conn_, initialCryptoStream);
-    auto& numProbePackets =
-        conn_->pendingEvents.numProbePackets[PacketNumberSpace::Initial];
-    if ((initialCryptoStream.retransmissionBuffer.size() &&
-         conn_->outstandings.packetCount[PacketNumberSpace::Initial] &&
-         numProbePackets) ||
-        initialScheduler.hasData() || toWriteInitialAcks(*conn_)) {
-      CHECK(conn_->initialHeaderCipher);
-      std::string& token = clientConn_->retryToken.empty()
-          ? clientConn_->newToken
-          : clientConn_->retryToken;
-      packetLimit -= writeCryptoAndAckDataToSocket(
-                         *socket_,
-                         *conn_,
-                         srcConnId /* src */,
-                         destConnId /* dst */,
-                         LongHeader::Types::Initial,
-                         *conn_->initialWriteCipher,
-                         *conn_->initialHeaderCipher,
-                         version,
-                         packetLimit,
-                         token)
-                         .packetsWritten;
-    }
+    const std::string& token = clientConn_->retryToken.empty()
+        ? clientConn_->newToken
+        : clientConn_->retryToken;
+    packetLimit -=
+        handleInitialWriteDataCommon(srcConnId, destConnId, packetLimit, token)
+            .packetsWritten;
     if (!packetLimit && !conn_->pendingEvents.anyProbePackets()) {
       return;
     }
   }
   if (conn_->handshakeWriteCipher) {
-    auto& handshakeCryptoStream =
-        *getCryptoStream(*conn_->cryptoState, EncryptionLevel::Handshake);
-    CryptoStreamScheduler handshakeScheduler(*conn_, handshakeCryptoStream);
-    auto& numProbePackets =
-        conn_->pendingEvents.numProbePackets[PacketNumberSpace::Handshake];
-    if ((conn_->outstandings.packetCount[PacketNumberSpace::Handshake] &&
-         handshakeCryptoStream.retransmissionBuffer.size() &&
-         numProbePackets) ||
-        handshakeScheduler.hasData() || toWriteHandshakeAcks(*conn_)) {
-      CHECK(conn_->handshakeWriteHeaderCipher);
-      packetLimit -= writeCryptoAndAckDataToSocket(
-                         *socket_,
-                         *conn_,
-                         srcConnId /* src */,
-                         destConnId /* dst */,
-                         LongHeader::Types::Handshake,
-                         *conn_->handshakeWriteCipher,
-                         *conn_->handshakeWriteHeaderCipher,
-                         version,
-                         packetLimit)
-                         .packetsWritten;
-    }
+    packetLimit -=
+        handleHandshakeWriteDataCommon(srcConnId, destConnId, packetLimit)
+            .packetsWritten;
     if (!packetLimit && !conn_->pendingEvents.anyProbePackets()) {
       return;
     }
@@ -1074,7 +1049,7 @@ bool QuicClientTransport::isTLSResumed() const {
 }
 
 void QuicClientTransport::errMessage(
-    FOLLY_MAYBE_UNUSED const cmsghdr& cmsg) noexcept {
+    [[maybe_unused]] const cmsghdr& cmsg) noexcept {
 #ifdef FOLLY_HAVE_MSG_ERRQUEUE
   if ((cmsg.cmsg_level == SOL_IP && cmsg.cmsg_type == IP_RECVERR) ||
       (cmsg.cmsg_level == SOL_IPV6 && cmsg.cmsg_type == IPV6_RECVERR)) {
@@ -1106,9 +1081,10 @@ void QuicClientTransport::errMessage(
     auto errStr = folly::errnoStr(serr->ee_errno);
     if (!happyEyeballsState.shouldWriteToFirstSocket &&
         !happyEyeballsState.shouldWriteToSecondSocket) {
-      runOnEvbAsync([errString = std::move(errStr)](auto self) {
-        auto quicError =
-            QuicError(QuicErrorCode(LocalErrorCode::CONNECT_FAILED), errString);
+      runOnEvbAsync([errString = std::move(errStr)](auto self) mutable {
+        auto quicError = QuicError(
+            QuicErrorCode(LocalErrorCode::CONNECT_FAILED),
+            std::move(errString));
         auto clientPtr = static_cast<QuicClientTransport*>(self.get());
         clientPtr->closeImpl(std::move(quicError), false, false);
       });
@@ -1182,12 +1158,14 @@ void QuicClientTransport::recvMsg(
     msg.msg_iovlen = 1;
 #ifdef FOLLY_HAVE_MSG_ERRQUEUE
     bool useGRO = sock.getGRO() > 0;
-    bool useTS = sock.getTimestamping() > 0;
+    bool useTs = sock.getTimestamping() > 0;
+    bool recvTos = sock.getRecvTos();
+    bool checkCmsgs = useGRO || useTs || recvTos;
     char control
         [QuicAsyncUDPSocket::ReadCallback::OnDataAvailableParams::kCmsgSpace] =
             {};
 
-    if (useGRO || useTS) {
+    if (checkCmsgs) {
       msg.msg_control = control;
       msg.msg_controllen = sizeof(control);
 
@@ -1219,7 +1197,7 @@ void QuicClientTransport::recvMsg(
       break;
     }
 #ifdef FOLLY_HAVE_MSG_ERRQUEUE
-    if (useGRO) {
+    if (checkCmsgs) {
       QuicAsyncUDPSocket::fromMsg(params, msg);
 
       // truncated
@@ -1258,21 +1236,26 @@ void QuicClientTransport::recvMsg(
 
           offset += params.gro;
           remaining -= params.gro;
-          networkData.addPacket(ReceivedUdpPacket(std::move(tmp)));
+          networkData.addPacket(ReceivedUdpPacket(
+              std::move(tmp), ReceivedUdpPacket::Timings{}, params.tos));
         } else {
           // do not clone the last packet
           // start at offset, use all the remaining data
           readBuffer->trimStart(offset);
           DCHECK_EQ(readBuffer->length(), remaining);
           remaining = 0;
-          networkData.addPacket(ReceivedUdpPacket(std::move(readBuffer)));
+          networkData.addPacket(ReceivedUdpPacket(
+              std::move(readBuffer), ReceivedUdpPacket::Timings{}, params.tos));
         }
       }
     } else {
-      networkData.addPacket(ReceivedUdpPacket(std::move(readBuffer)));
+      networkData.addPacket(ReceivedUdpPacket(
+          std::move(readBuffer), ReceivedUdpPacket::Timings{}, params.tos));
     }
-    trackDatagramReceived(bytesRead);
+    maybeQlogDatagram(bytesRead);
   }
+  trackDatagramsReceived(
+      networkData.getPackets().size(), networkData.getTotalData());
 }
 
 void QuicClientTransport::recvMmsg(
@@ -1286,11 +1269,13 @@ void QuicClientTransport::recvMmsg(
   int flags = 0;
 #ifdef FOLLY_HAVE_MSG_ERRQUEUE
   bool useGRO = sock.getGRO() > 0;
-  bool useTS = sock.getTimestamping() > 0;
+  bool useTs = sock.getTimestamping() > 0;
+  bool recvTos = sock.getRecvTos();
+  bool checkCmsgs = useGRO || useTs || recvTos;
   std::vector<std::array<
       char,
       QuicAsyncUDPSocket::ReadCallback::OnDataAvailableParams::kCmsgSpace>>
-      controlVec((useGRO || useTS) ? numPackets : 0);
+      controlVec(checkCmsgs ? numPackets : 0);
 
   // we need to consider MSG_TRUNC too
   if (useGRO) {
@@ -1317,7 +1302,7 @@ void QuicClientTransport::recvMmsg(
     msg->msg_name = rawAddr;
     msg->msg_namelen = kAddrLen;
 #ifdef FOLLY_HAVE_MSG_ERRQUEUE
-    if (useGRO || useTS) {
+    if (checkCmsgs) {
       ::memset(controlVec[i].data(), 0, controlVec[i].size());
       msg->msg_control = controlVec[i].data();
       msg->msg_controllen = controlVec[i].size();
@@ -1360,7 +1345,7 @@ void QuicClientTransport::recvMmsg(
     }
     QuicAsyncUDPSocket::ReadCallback::OnDataAvailableParams params;
 #ifdef FOLLY_HAVE_MSG_ERRQUEUE
-    if (useGRO || useTS) {
+    if (checkCmsgs) {
       QuicAsyncUDPSocket::fromMsg(params, msg.msg_hdr);
 
       // truncated
@@ -1401,26 +1386,32 @@ void QuicClientTransport::recvMmsg(
 
           offset += params.gro;
           remaining -= params.gro;
-          networkData.addPacket(ReceivedUdpPacket(std::move(tmp)));
+          networkData.addPacket(ReceivedUdpPacket(
+              std::move(tmp), ReceivedUdpPacket::Timings{}, params.tos));
         } else {
           // do not clone the last packet
           // start at offset, use all the remaining data
           readBuffer->trimStart(offset);
           DCHECK_EQ(readBuffer->length(), remaining);
           remaining = 0;
-          networkData.addPacket(ReceivedUdpPacket(std::move(readBuffer)));
+          networkData.addPacket(ReceivedUdpPacket(
+              std::move(readBuffer), ReceivedUdpPacket::Timings{}, params.tos));
         }
       }
     } else {
-      networkData.addPacket(ReceivedUdpPacket(std::move(readBuffer)));
+      networkData.addPacket(ReceivedUdpPacket(
+          std::move(readBuffer), ReceivedUdpPacket::Timings{}, params.tos));
     }
 
-    trackDatagramReceived(bytesRead);
+    maybeQlogDatagram(bytesRead);
   }
+  trackDatagramsReceived(
+      networkData.getPackets().size(), networkData.getTotalData());
 }
 
 void QuicClientTransport::onNotifyDataAvailable(
     QuicAsyncUDPSocket& sock) noexcept {
+  auto self = this->shared_from_this();
   CHECK(conn_) << "trying to receive packets without a connection";
   auto readBufferSize =
       conn_->transportSettings.maxRecvPacketSize * numGROBuffers_;
@@ -1437,11 +1428,14 @@ void QuicClientTransport::onNotifyDataAvailable(
 
     // track the received packets
     for (const auto& packet : networkData.getPackets()) {
-      if (!packet.buf) {
+      if (packet.buf.empty()) {
         continue;
       }
-      trackDatagramReceived(packet.buf->computeChainDataLength());
+      auto len = packet.buf.chainLength();
+      maybeQlogDatagram(len);
     }
+    trackDatagramsReceived(
+        networkData.getPackets().size(), networkData.getTotalData());
 
     // Propagate errors
     // TODO(bschlinker): Investigate generalization of loopDetectorCallback
@@ -1550,6 +1544,7 @@ void QuicClientTransport::start(
         conn_->localAddress,
         conn_->peerAddress,
         conn_->transportSettings,
+        conn_->socketTos.value,
         this,
         this,
         socketOptions_);
@@ -1678,6 +1673,7 @@ void QuicClientTransport::onNetworkSwitch(
         conn_->localAddress,
         conn_->peerAddress,
         conn_->transportSettings,
+        conn_->socketTos.value,
         this,
         this,
         socketOptions_);
@@ -1696,17 +1692,23 @@ void QuicClientTransport::setTransportStatsCallback(
   statsCallback_ = std::move(statsCallback);
   if (statsCallback_) {
     conn_->statsCallback = statsCallback_.get();
+    conn_->readCodec->setConnectionStatsCallback(statsCallback_.get());
   } else {
     conn_->statsCallback = nullptr;
   }
 }
 
-void QuicClientTransport::trackDatagramReceived(size_t len) {
+void QuicClientTransport::maybeQlogDatagram(size_t len) {
   if (conn_->qLogger) {
     conn_->qLogger->addDatagramReceived(len);
   }
-  QUIC_STATS(statsCallback_, onPacketReceived);
-  QUIC_STATS(statsCallback_, onRead, len);
+}
+
+void QuicClientTransport::trackDatagramsReceived(
+    uint32_t totalPackets,
+    uint32_t totalPacketLen) {
+  QUIC_STATS(statsCallback_, onPacketsReceived, totalPackets);
+  QUIC_STATS(statsCallback_, onRead, totalPacketLen);
 }
 
 void QuicClientTransport::maybeSendTransportKnobs() {
@@ -1726,6 +1728,31 @@ void QuicClientTransport::maybeSendTransportKnobs() {
   }
 }
 
+folly::Optional<std::vector<TransportParameter>>
+QuicClientTransport::getPeerTransportParams() const {
+  if (clientConn_ && clientConn_->clientHandshakeLayer) {
+    auto maybeParams =
+        clientConn_->clientHandshakeLayer->getServerTransportParams();
+    if (maybeParams) {
+      return maybeParams->parameters;
+    }
+  }
+  return folly::none;
+}
+
+void QuicClientTransport::setCongestionControl(CongestionControlType type) {
+  if (!conn_->congestionControllerFactory) {
+    // If you are hitting this, update your application to call
+    // setCongestionControllerFactory() on the transport and share one factory
+    // for all transports.
+    conn_->congestionControllerFactory =
+        std::make_shared<DefaultCongestionControllerFactory>();
+    LOG(WARNING)
+        << "A congestion controller factory is not set. Using a default per-transport instance.";
+  }
+  QuicTransportBase::setCongestionControl(type);
+}
+
 void QuicClientTransport::RecvmmsgStorage::resize(size_t numPackets) {
   if (msgs.size() != numPackets) {
     msgs.resize(numPackets);
