@@ -45,7 +45,7 @@ enum class TokenType : uint8_t { RetryToken = 0, NewToken };
 
 constexpr uint8_t kHeaderFormMask = 0x80;
 constexpr auto kMaxPacketNumEncodingSize = 4;
-constexpr auto kNumInitialAckBlocksPerFrame = 32;
+constexpr auto kNumInitialAckBlocksPerFrame = 3;
 
 template <class T>
 using IntervalSetVec = SmallVec<T, kNumInitialAckBlocksPerFrame>;
@@ -190,6 +190,9 @@ struct ReadAckFrame {
   folly::Optional<std::chrono::microseconds> maybeLatestRecvdPacketTime;
   folly::Optional<PacketNum> maybeLatestRecvdPacketNum;
   RecvdPacketsTimestampsRangeVec recvdPacketsTimestampRanges;
+  uint32_t ecnECT0Count{0};
+  uint32_t ecnECT1Count{0};
+  uint32_t ecnCECount{0};
   bool operator==(const ReadAckFrame& /*rhs*/) const {
     // Can't compare ackBlocks, function is just here to appease compiler.
     return false;
@@ -208,6 +211,9 @@ struct WriteAckFrame {
   folly::Optional<std::chrono::microseconds> maybeLatestRecvdPacketTime;
   folly::Optional<PacketNum> maybeLatestRecvdPacketNum;
   RecvdPacketsTimestampsRangeVec recvdPacketsTimestampRanges;
+  uint32_t ecnECT0Count{0};
+  uint32_t ecnECT1Count{0};
+  uint32_t ecnCECount{0};
   bool operator==(const WriteAckFrame& /*rhs*/) const {
     // Can't compare ackBlocks, function is just here to appease compiler.
     return false;
@@ -251,6 +257,10 @@ struct WriteAckFrameState {
   // element in the deque (e.g., entries are not added for packets that
   // arrive out of order relative to previously received packets).
   CircularDeque<ReceivedPacket> recvdPacketInfos;
+  // The count of ECN marks seen on received packets.
+  uint32_t ecnECT0CountReceived{0};
+  uint32_t ecnECT1CountReceived{0};
+  uint32_t ecnCECountReceived{0};
 };
 
 struct WriteAckFrameMetaData {
@@ -1155,16 +1165,17 @@ struct RegularPacket {
 };
 
 struct RetryPacket {
+  using IntegrityTagType = std::array<uint8_t, kRetryIntegrityTagLen>;
   RetryPacket(
       LongHeader&& longHeaderIn,
-      Buf integrityTagIn,
+      IntegrityTagType integrityTagIn,
       uint8_t initialByteIn)
       : header(std::move(longHeaderIn)),
-        integrityTag(std::move(integrityTagIn)),
+        integrityTag(integrityTagIn),
         initialByte(initialByteIn) {}
 
   LongHeader header;
-  Buf integrityTag;
+  std::array<uint8_t, kRetryIntegrityTagLen> integrityTag;
   uint8_t initialByte;
 };
 
@@ -1195,6 +1206,15 @@ struct RegularQuicWritePacket : public RegularPacket {
       : RegularPacket(std::move(headerIn)) {}
 };
 
+enum class ECNState : uint8_t {
+  NotAttempted, // ECN marking disabled
+  AttemptingECN, // ECN enabled, but not yet negotiated
+  ValidatedECN, // ECN enabled, and negotiated successfully
+  AttemptingL4S, // L4S enabled, but not yet negotiated
+  ValidatedL4S, // L4S enabled, and negotiated successfully
+  FailedValidation, // ECN or L4S was enabled, but failed validation
+};
+
 /**
  * Returns whether the header is long or short from the initial byte of
  * the QUIC packet.
@@ -1203,9 +1223,7 @@ struct RegularQuicWritePacket : public RegularPacket {
  */
 HeaderForm getHeaderForm(uint8_t headerValue);
 
-std::string toString(LongHeader::Types type);
-
-std::string toString(QuicErrorCode code);
+std::string_view toString(LongHeader::Types type);
 
 inline std::ostream& operator<<(
     std::ostream& os,
@@ -1228,18 +1246,18 @@ inline std::ostream& operator<<(std::ostream& os, const PacketHeader& header) {
   return os;
 }
 
-std::string toString(PacketNumberSpace pnSpace);
+std::string_view toString(PacketNumberSpace pnSpace);
 
-std::string toString(FrameType frame);
+std::string_view toString(FrameType frame);
 
-std::string toString(QuicVersion version);
+std::string_view toString(QuicVersion version);
 
 inline std::ostream& operator<<(std::ostream& os, PacketNumberSpace pnSpace) {
   return os << toString(pnSpace);
 }
 
-std::string toString(ProtectionType protectionType);
+std::string_view toString(ProtectionType protectionType);
 
-std::string toString(TokenType type);
+std::string_view toString(TokenType type);
 
 } // namespace quic
