@@ -7,8 +7,8 @@
 
 #include <quic/common/test/TestUtils.h>
 
+#include <fizz/backend/openssl/certificate/OpenSSLSelfCertImpl.h>
 #include <fizz/crypto/test/TestUtil.h>
-#include <fizz/protocol/OpenSSLSelfCertImpl.h>
 #include <fizz/protocol/clock/test/Mocks.h>
 #include <fizz/protocol/test/Mocks.h>
 #include <quic/api/QuicTransportFunctions.h>
@@ -116,19 +116,28 @@ static std::shared_ptr<fizz::SelfCert> readCert() {
   auto privKey = fizz::test::getPrivateKey(fizz::test::kP256Key);
   std::vector<folly::ssl::X509UniquePtr> certs;
   certs.emplace_back(std::move(certificate));
-  return std::make_shared<fizz::OpenSSLSelfCertImpl<fizz::KeyType::P256>>(
+  return std::make_shared<
+      fizz::openssl::OpenSSLSelfCertImpl<fizz::openssl::KeyType::P256>>(
       std::move(privKey), std::move(certs));
 }
 
+std::shared_ptr<fizz::client::FizzClientContext> createClientCtx() {
+  auto clientCtx = std::make_shared<fizz::client::FizzClientContext>();
+  clientCtx->setClock(std::make_shared<NiceMock<fizz::test::MockClock>>());
+  clientCtx->setSupportedAlpns({"quic_test"});
+  return clientCtx;
+}
+
 std::shared_ptr<fizz::server::FizzServerContext> createServerCtx() {
   auto cert = readCert();
   auto certManager = std::make_unique<fizz::server::CertManager>();
-  certManager->addCert(std::move(cert), true);
+  certManager->addCertAndSetDefault(std::move(cert));
   auto serverCtx = std::make_shared<fizz::server::FizzServerContext>();
   serverCtx->setFactory(std::make_shared<QuicFizzFactory>());
   serverCtx->setCertManager(std::move(certManager));
   serverCtx->setOmitEarlyRecordLayer(true);
   serverCtx->setClock(std::make_shared<NiceMock<fizz::test::MockClock>>());
+  serverCtx->setSupportedAlpns({"quic_test"});
   return serverCtx;
 }
 
@@ -239,7 +248,7 @@ QuicCachedPsk setupZeroRttOnClientCtx(
 void setupCtxWithTestCert(fizz::server::FizzServerContext& ctx) {
   auto cert = readCert();
   auto certManager = std::make_unique<fizz::server::CertManager>();
-  certManager->addCert(std::move(cert), true);
+  certManager->addCertAndSetDefault(std::move(cert));
   ctx.setCertManager(std::move(certManager));
 }
 
@@ -383,42 +392,48 @@ RegularQuicPacketBuilder::Packet createCryptoPacket(
 }
 
 Buf packetToBuf(const RegularQuicPacketBuilder::Packet& packet) {
-  auto packetBuf = packet.header->clone();
-  if (packet.body) {
-    packetBuf->prependChain(packet.body->clone());
+  auto packetBuf = packet.header.clone();
+  if (!packet.body.empty()) {
+    packetBuf->prependChain(packet.body.clone());
   }
   return packetBuf;
 }
 
+ReceivedUdpPacket packetToReceivedUdpPacket(
+    const RegularQuicPacketBuilder::Packet& writePacket) {
+  ReceivedUdpPacket packet(packetToBuf(writePacket));
+  return packet;
+}
+
 Buf packetToBufCleartext(
-    const RegularQuicPacketBuilder::Packet& packet,
+    RegularQuicPacketBuilder::Packet& packet,
     const Aead& cleartextCipher,
     const PacketNumberCipher& headerCipher,
     PacketNum packetNum) {
   VLOG(10) << __func__ << " packet header: "
-           << folly::hexlify(packet.header->clone()->moveToFbString());
-  auto packetBuf = packet.header->clone();
+           << folly::hexlify(packet.header.clone()->moveToFbString());
+  auto packetBuf = packet.header.clone();
   Buf body;
-  if (packet.body) {
-    packet.body->coalesce();
-    body = packet.body->clone();
+  if (!packet.body.empty()) {
+    packet.body.coalesce();
+    body = packet.body.clone();
   } else {
     body = folly::IOBuf::create(0);
   }
   auto headerForm = packet.packet.header.getHeaderForm();
-  packet.header->coalesce();
+  packet.header.coalesce();
   auto tagLen = cleartextCipher.getCipherOverhead();
   if (body->tailroom() < tagLen) {
     body->prependChain(folly::IOBuf::create(tagLen));
   }
   body->coalesce();
   auto encryptedBody = cleartextCipher.inplaceEncrypt(
-      std::move(body), packet.header.get(), packetNum);
+      std::move(body), &packet.header, packetNum);
   encryptedBody->coalesce();
   encryptPacketHeader(
       headerForm,
-      packet.header->writableData(),
-      packet.header->length(),
+      packet.header.writableData(),
+      packet.header.length(),
       encryptedBody->data(),
       encryptedBody->length(),
       headerCipher);
@@ -426,6 +441,14 @@ Buf packetToBufCleartext(
   return packetBuf;
 }
 
+Buf packetToBufCleartext(
+    RegularQuicPacketBuilder::Packet&& packet,
+    const Aead& cleartextCipher,
+    const PacketNumberCipher& headerCipher,
+    PacketNum packetNum) {
+  return packetToBufCleartext(packet, cleartextCipher, headerCipher, packetNum);
+}
+
 uint64_t computeExpectedDelay(
     std::chrono::microseconds ackDelay,
     uint8_t ackDelayExponent) {
@@ -480,10 +503,10 @@ void updateAckState(
     bool pkHasRetransmittableData,
     bool pkHasCryptoData,
     TimePoint receiveTimePoint) {
-  ReceivedUdpPacket::Timings packetTimings;
-  packetTimings.receiveTimePoint = receiveTimePoint;
-  uint64_t distance = addPacketToAckState(
-      conn, getAckState(conn, pnSpace), packetNum, packetTimings);
+  ReceivedUdpPacket packet;
+  packet.timings.receiveTimePoint = receiveTimePoint;
+  uint64_t distance =
+      addPacketToAckState(conn, getAckState(conn, pnSpace), packetNum, packet);
   updateAckSendStateOnRecvPacket(
       conn,
       getAckState(conn, pnSpace),
@@ -528,9 +551,7 @@ OutstandingPacketWrapper makeTestingWritePacket(
       0,
       false,
       totalBytesSent,
-      0,
       inflightBytes,
-      0,
       LossState(),
       writeCount,
       OutstandingPacketMetadata::DetailsPerStream());
@@ -554,22 +575,21 @@ CongestionController::AckEvent makeAck(
 
   ack.ackedBytes = ackedSize;
   ack.largestNewlyAckedPacket = seq;
+  OutstandingPacketMetadata opm(
+      sentTime,
+      ackedSize /* encodedSize */,
+      ackedSize /* encodedBodySize */,
+      false /* isHandshake */,
+      0 /* totalBytesSent */,
+      0 /* inflightBytes */,
+      LossState() /* lossState */,
+      0 /* writeCount */,
+      OutstandingPacketMetadata::DetailsPerStream());
   ack.ackedPackets.emplace_back(
       CongestionController::AckEvent::AckPacket::Builder()
           .setPacketNum(seq)
           .setNonDsrPacketSequenceNumber(seq)
-          .setOutstandingPacketMetadata(OutstandingPacketMetadata(
-              sentTime,
-              ackedSize /* encodedSize */,
-              ackedSize /* encodedBodySize */,
-              false /* isHandshake */,
-              0 /* totalBytesSent */,
-              0 /* totalBodyBytesSent */,
-              0 /* inflightBytes */,
-              0 /* numOutstanding */,
-              LossState() /* lossState */,
-              0 /* writeCount */,
-              OutstandingPacketMetadata::DetailsPerStream()))
+          .setOutstandingPacketMetadata(opm)
           .setDetailsPerStream(AckEvent::AckPacket::DetailsPerStream())
           .build());
   ack.largestNewlyAckedPacketSentTime = sentTime;
@@ -706,8 +726,11 @@ CongestionController::AckEvent::AckPacket makeAckPacketFromOutstandingPacket(
       .setPacketNum(outstandingPacket.packet.header.getPacketSequenceNum())
       .setNonDsrPacketSequenceNumber(
           outstandingPacket.packet.header.getPacketSequenceNum())
-      .setOutstandingPacketMetadata(std::move(outstandingPacket.metadata))
-      .setLastAckedPacketInfo(std::move(outstandingPacket.lastAckedPacketInfo))
+      .setOutstandingPacketMetadata(outstandingPacket.metadata)
+      .setLastAckedPacketInfo(
+          outstandingPacket.lastAckedPacketInfo
+              ? &*outstandingPacket.lastAckedPacketInfo
+              : nullptr)
       .setAppLimited(outstandingPacket.isAppLimited)
       .setDetailsPerStream(
           CongestionController::AckEvent::AckPacket::DetailsPerStream())
@@ -723,7 +746,7 @@ writeCryptoFrame(uint64_t offsetIn, Buf data, PacketBuilderInterface& builder) {
 void overridePacketWithToken(
     PacketBuilderInterface::Packet& packet,
     const StatelessResetToken& token) {
-  overridePacketWithToken(*packet.body, token);
+  overridePacketWithToken(packet.body, token);
 }
 
 void overridePacketWithToken(
