@@ -26,7 +26,6 @@
 #include <fizz/server/FizzServerContext.h>
 #include <quic/common/testutil/MockAsyncUDPSocket.h>
 
-#include <folly/ssl/Init.h>
 #include <quic/codec/QuicConnectionId.h>
 
 namespace quic {
@@ -118,8 +117,17 @@ RegularQuicPacketBuilder::Packet createCryptoPacket(
 
 Buf packetToBuf(const RegularQuicPacketBuilder::Packet& packet);
 
+ReceivedUdpPacket packetToReceivedUdpPacket(
+    const RegularQuicPacketBuilder::Packet& packetIn);
+
 Buf packetToBufCleartext(
-    const RegularQuicPacketBuilder::Packet& packet,
+    RegularQuicPacketBuilder::Packet& packet,
+    const Aead& cleartextCipher,
+    const PacketNumberCipher& headerCipher,
+    PacketNum packetNum);
+
+Buf packetToBufCleartext(
+    RegularQuicPacketBuilder::Packet&& packet,
     const Aead& cleartextCipher,
     const PacketNumberCipher& headerCipher,
     PacketNum packetNum);
@@ -132,6 +140,8 @@ bool isState(const S& s) {
       [](const auto&) { return false; });
 }
 
+std::shared_ptr<fizz::client::FizzClientContext> createClientCtx();
+
 std::shared_ptr<fizz::server::FizzServerContext> createServerCtx();
 
 void setupCtxWithTestCert(fizz::server::FizzServerContext& ctx);
@@ -527,6 +537,16 @@ class FakeServerHandshake : public FizzServerHandshake {
     }
     oneRttReadCipher_ = createNoOpAead();
     oneRttReadHeaderCipher_ = createNoOpHeaderCipher();
+    readTrafficSecret_ = folly::IOBuf::copyBuffer(getRandSecret());
+    writeTrafficSecret_ = folly::IOBuf::copyBuffer(getRandSecret());
+  }
+
+  std::unique_ptr<Aead> buildAead(folly::ByteRange /*secret*/) override {
+    return createNoOpAead();
+  }
+
+  Buf getNextTrafficSecret(folly::ByteRange /*secret*/) const override {
+    return folly::IOBuf::copyBuffer(getRandSecret());
   }
 
   void setHandshakeKeys() {
