@@ -26,13 +26,13 @@ constexpr uint64_t kMaxExtraAckedFilterLen =
 constexpr float kStartupPacingGain = 2.89; // 2 / ln(2)
 constexpr float kDrainPacingGain = 1 / kStartupPacingGain;
 constexpr float kProbeBwDownPacingGain = 0.9;
-constexpr float kProbeBwCruisePacingGain = 1.0;
-constexpr float kProbeBwRefillPacingGain = 1.0;
+constexpr float kProbeBwCruiseRefillPacingGain = 1.0;
 constexpr float kProbeBwUpPacingGain = 1.25;
 constexpr float kProbeRttPacingGain = 1.0;
 
 constexpr float kStartupCwndGain = 2.89;
-constexpr float kProbeBwCwndGain = 2.0;
+constexpr float kProbeBwCruiseRefillCwndGain = 2.0;
+constexpr float kProbeBwUpDownCwndGain = 2.0;
 constexpr float kProbeRttCwndGain = 0.5;
 
 constexpr float kBeta = 0.7;
@@ -41,7 +41,10 @@ constexpr float kLossThreshold = 0.02;
 constexpr float kHeadroomFactor = 0.15;
 
 quic::Bandwidth kMinPacingRateForSendQuantum{1200 * 1000, 1s};
-constexpr uint8_t kPacingMarginPercent = 1;
+// The experimental pacer currently achieves ~99% of the target rate
+// we should not reduce the target by adding an extra margin.
+// TODO: add the margin back if the pacer performance improves further.
+constexpr uint8_t kPacingMarginPercent = 0;
 
 Bbr2CongestionController::Bbr2CongestionController(
     QuicConnectionStateBase& conn)
@@ -84,8 +87,9 @@ void Bbr2CongestionController::onPacketSent(
   addAndCheckOverflow(
       conn_.lossState.inflightBytes, packet.metadata.encodedSize);
 
-  // Maintain cwndLimited flag
-  if (conn_.lossState.inflightBytes >= cwndBytes_) {
+  // Maintain cwndLimited flag. We consider the transport being cwnd limited if
+  // we are using > 90% of the cwnd.
+  if (conn_.lossState.inflightBytes > cwndBytes_ * 9 / 10) {
     cwndLimitedInRound_ |= true;
   }
 }
@@ -115,7 +119,8 @@ void Bbr2CongestionController::onPacketAckOrLoss(
             << ")";
   };
 
-  if (lossEvent && lossEvent->lostPackets > 0) {
+  if (lossEvent && lossEvent->lostPackets > 0 &&
+      conn_.transportSettings.ccaConfig.conservativeRecovery) {
     // The pseudo code in BBRHandleLostPacket is included in
     // updateProbeBwCyclePhase. No need to repeat it here.
 
@@ -123,6 +128,9 @@ void Bbr2CongestionController::onPacketAckOrLoss(
     // non-persistent congestion
     saveCwnd();
     inPacketConservation_ = true;
+    // Mark the connection as app-limited so bw samples during recovery are not
+    // taken into account.
+    setAppLimited();
     packetConservationStartTime_ = Clock::now();
     if (lossEvent->persistentCongestion) {
       cwndBytes_ = kMinCwndInMssForBbr * conn_.udpSendPacketLen;
@@ -134,9 +142,18 @@ void Bbr2CongestionController::onPacketAckOrLoss(
   }
 
   if (ackEvent) {
+    if (ackEvent->implicit) {
+      // Implicit acks should not be used for bandwidth or rtt estimation
+      setCwnd(ackEvent->ackedBytes, 0);
+      return;
+    }
+
     if (appLimited_ &&
         appLimitedLastSendTime_ <= ackEvent->largestNewlyAckedPacketSentTime) {
       appLimited_ = false;
+      if (conn_.qLogger) {
+        conn_.qLogger->addAppUnlimitedUpdate();
+      }
     }
 
     if (inPacketConservation_ &&
@@ -209,6 +226,9 @@ bool Bbr2CongestionController::isAppLimited() const {
 void Bbr2CongestionController::setAppLimited() noexcept {
   appLimited_ = true;
   appLimitedLastSendTime_ = Clock::now();
+  if (conn_.qLogger) {
+    conn_.qLogger->addAppLimitedUpdate();
+  }
 }
 
 // Internals
@@ -225,8 +245,7 @@ void Bbr2CongestionController::resetLowerBounds() {
 }
 void Bbr2CongestionController::enterStartup() {
   state_ = State::Startup;
-  pacingGain_ = kStartupPacingGain;
-  cwndGain_ = kStartupCwndGain;
+  updatePacingAndCwndGain();
 }
 
 void Bbr2CongestionController::setPacing() {
@@ -265,7 +284,7 @@ void Bbr2CongestionController::setCwnd(
   auto inflightMax = addQuantizationBudget(
       getBDPWithGain(cwndGain_) + maxExtraAckedFilter_.GetBest());
   // BBRModulateCwndForRecovery()
-  if (lostBytes > 0) {
+  if (lostBytes > 0 && !conn_.transportSettings.ccaConfig.ignoreLoss) {
     cwndBytes_ = std::max(
         cwndBytes_ - std::min(lostBytes, cwndBytes_),
         kMinCwndInMssForBbr * conn_.udpSendPacketLen);
@@ -313,7 +332,8 @@ void Bbr2CongestionController::setCwnd(
 
 void Bbr2CongestionController::checkProbeRttDone() {
   auto timeNow = Clock::now();
-  if (probeRttDoneTimestamp_ && timeNow > *probeRttDoneTimestamp_) {
+  if ((probeRttDoneTimestamp_ && timeNow > *probeRttDoneTimestamp_) ||
+      conn_.lossState.inflightBytes == 0) {
     // Schedule the next ProbeRTT
     probeRttMinTimestamp_ = timeNow;
     restoreCwnd();
@@ -327,7 +347,6 @@ void Bbr2CongestionController::restoreCwnd() {
 }
 void Bbr2CongestionController::exitProbeRtt() {
   resetLowerBounds();
-  cwndGain_ = kProbeBwCwndGain;
   if (filledPipe_) {
     startProbeBwDown();
     startProbeBwCruise();
@@ -383,7 +402,7 @@ void Bbr2CongestionController::updateCongestionSignals(
   if (state_ == State::ProbeBw_Up) {
     return;
   }
-  if (lossBytesInRound_ > 0 && !conn_.transportSettings.ccaConfig.ignoreLoss) {
+  if (lossBytesInRound_ > 0) {
     // InitLowerBounds
     if (!bandwidthLo_.has_value()) {
       bandwidthLo_ = maxBwFilter_.GetBest();
@@ -424,9 +443,6 @@ void Bbr2CongestionController::checkStartupDone() {
   checkStartupHighLoss();
 
   if (state_ == State::Startup && filledPipe_) {
-    if (conn_.transportSettings.ccaConfig.advanceCycleAfterStartup) {
-      cycleCount_++;
-    }
     enterDrain();
   }
 }
@@ -477,8 +493,7 @@ void Bbr2CongestionController::checkStartupHighLoss() {
 
 void Bbr2CongestionController::enterDrain() {
   state_ = State::Drain;
-  pacingGain_ = kDrainPacingGain; // Slow down pacing
-  cwndGain_ = kStartupCwndGain; // maintain cwnd
+  updatePacingAndCwndGain();
 }
 
 void Bbr2CongestionController::checkDrain() {
@@ -526,9 +541,8 @@ void Bbr2CongestionController::updateProbeBwCyclePhase(
       }
       break;
     case State::ProbeBw_Up:
-
       if (hasElapsedInPhase(minRtt_) &&
-          inflightLatest_ > getTargetInflightWithGain(1.25)) {
+          conn_.lossState.inflightBytes > getTargetInflightWithGain(1.25)) {
         startProbeBwDown();
       }
       break;
@@ -544,14 +558,7 @@ void Bbr2CongestionController::adaptUpperBounds(
     uint64_t ackedBytes,
     uint64_t inflightBytesAtLargestAckedPacket,
     uint64_t lostBytes) {
-  /* Advance probeBw round if necessary and update BBR.max_bw window and
-   * BBR.inflight_hi and BBR.bw_hi. */
-  if (state_ == State::ProbeBw_Down && roundStart_) {
-    /* end of samples from bw probing phase */
-    if (!isAppLimited()) {
-      cycleCount_++;
-    }
-  }
+  /* Update BBR.inflight_hi and BBR.bw_hi. */
 
   if (!checkInflightTooHigh(inflightBytesAtLargestAckedPacket, lostBytes)) {
     if (!inflightHi_.has_value() || !bandwidthHi_.has_value()) {
@@ -708,8 +715,7 @@ void Bbr2CongestionController::checkProbeRtt(uint64_t ackedBytes) {
 
 void Bbr2CongestionController::enterProbeRtt() {
   state_ = State::ProbeRTT;
-  pacingGain_ = kProbeRttPacingGain;
-  cwndGain_ = kProbeRttCwndGain;
+  updatePacingAndCwndGain();
 }
 
 void Bbr2CongestionController::handleProbeRtt() {
@@ -752,16 +758,19 @@ void Bbr2CongestionController::boundCwndForProbeRTT() {
 }
 
 void Bbr2CongestionController::boundBwForModel() {
+  Bandwidth previousBw = bandwidth_;
   bandwidth_ = maxBwFilter_.GetBest();
   if (state_ != State::Startup) {
-    if (bandwidthLo_.has_value()) {
+    if (bandwidthLo_.has_value() &&
+        !conn_.transportSettings.ccaConfig.ignoreLoss) {
       bandwidth_ = std::min(bandwidth_, *bandwidthLo_);
     }
-    if (bandwidthHi_.has_value()) {
+    if (bandwidthHi_.has_value() &&
+        !conn_.transportSettings.ccaConfig.ignoreInflightHi) {
       bandwidth_ = std::min(bandwidth_, *bandwidthHi_);
     }
   }
-  if (conn_.qLogger) {
+  if (conn_.qLogger && previousBw != bandwidth_) {
     conn_.qLogger->addBandwidthEstUpdate(bandwidth_.units, bandwidth_.interval);
   }
 }
@@ -802,7 +811,6 @@ uint64_t Bbr2CongestionController::getBDPWithGain(float gain) const {
 }
 
 void Bbr2CongestionController::enterProbeBW() {
-  cwndGain_ = kProbeBwCwndGain;
   startProbeBwDown();
 }
 
@@ -829,16 +837,23 @@ void Bbr2CongestionController::startProbeBwDown() {
   /* Decide random round-trip bound for wait: */
   roundsSinceBwProbe_ = folly::Random::rand32() % 2;
   /* Decide the random wall clock bound for wait: between 2-3 seconds */
-  bwProbeWait_ = std::chrono::milliseconds(2 + folly::Random::rand32() % 1000);
+  bwProbeWait_ =
+      std::chrono::milliseconds(2000 + (folly::Random::rand32() % 1000));
 
   probeBWCycleStart_ = Clock::now();
   state_ = State::ProbeBw_Down;
-  pacingGain_ = kProbeBwDownPacingGain;
+  updatePacingAndCwndGain();
   startRound();
+
+  // This is a new ProbeBW cycle. Advance the max bw filter if we're not app
+  // limited
+  if (!isAppLimited()) {
+    cycleCount_++;
+  }
 }
 void Bbr2CongestionController::startProbeBwCruise() {
   state_ = State::ProbeBw_Cruise;
-  pacingGain_ = kProbeBwCruisePacingGain;
+  updatePacingAndCwndGain();
 }
 
 void Bbr2CongestionController::startProbeBwRefill() {
@@ -846,13 +861,13 @@ void Bbr2CongestionController::startProbeBwRefill() {
   probeUpRounds_ = 0;
   probeUpAcks_ = 0;
   state_ = State::ProbeBw_Refill;
-  pacingGain_ = kProbeBwRefillPacingGain;
+  updatePacingAndCwndGain();
   startRound();
 }
 void Bbr2CongestionController::startProbeBwUp() {
   probeBWCycleStart_ = Clock::now();
   state_ = State::ProbeBw_Up;
-  pacingGain_ = kProbeBwUpPacingGain;
+  updatePacingAndCwndGain();
   startRound();
   raiseInflightHiSlope();
 }
@@ -911,6 +926,9 @@ Bandwidth Bbr2CongestionController::getBandwidthSampleFromAck(
 }
 
 bool Bbr2CongestionController::isRenoCoexistenceProbeTime() {
+  if (!conn_.transportSettings.ccaConfig.enableRenoCoexistence) {
+    return false;
+  }
   auto renoBdpInPackets = std::min(getTargetInflightWithGain(), cwndBytes_) /
       conn_.udpSendPacketLen;
   auto roundsBeforeRenoProbe =
@@ -928,6 +946,44 @@ void Bbr2CongestionController::getStats(
   stats.bbr2Stats.state = uint8_t(state_);
 }
 
+void Bbr2CongestionController::updatePacingAndCwndGain() {
+  switch (state_) {
+    case State::Startup:
+      pacingGain_ =
+          conn_.transportSettings.ccaConfig.overrideStartupPacingGain > 0
+          ? conn_.transportSettings.ccaConfig.overrideStartupPacingGain
+          : kStartupPacingGain;
+      cwndGain_ = kStartupCwndGain;
+      break;
+    case State::Drain:
+      pacingGain_ = kDrainPacingGain;
+      cwndGain_ = kStartupCwndGain;
+      break;
+    case State::ProbeBw_Up:
+      pacingGain_ = kProbeBwUpPacingGain;
+      cwndGain_ = kProbeBwUpDownCwndGain;
+      break;
+    case State::ProbeBw_Down:
+      pacingGain_ = kProbeBwDownPacingGain;
+      cwndGain_ = kProbeBwUpDownCwndGain;
+      break;
+    case State::ProbeBw_Cruise:
+    case State::ProbeBw_Refill:
+      pacingGain_ =
+          conn_.transportSettings.ccaConfig.overrideCruisePacingGain > 0
+          ? conn_.transportSettings.ccaConfig.overrideCruisePacingGain
+          : kProbeBwCruiseRefillPacingGain;
+      cwndGain_ = conn_.transportSettings.ccaConfig.overrideCruiseCwndGain > 0
+          ? conn_.transportSettings.ccaConfig.overrideCruiseCwndGain
+          : kProbeBwCruiseRefillCwndGain;
+      break;
+    case State::ProbeRTT:
+      pacingGain_ = kProbeRttPacingGain;
+      cwndGain_ = kProbeRttCwndGain;
+      break;
+  }
+}
+
 std::string bbr2StateToString(Bbr2CongestionController::State state) {
   switch (state) {
     case Bbr2CongestionController::State::Startup:
@@ -947,4 +1003,5 @@ std::string bbr2StateToString(Bbr2CongestionController::State state) {
   }
   folly::assume_unreachable();
 }
+
 } // namespace quic
