@@ -23,7 +23,6 @@
 #include <folly/io/async/SSLContext.h>
 #include <folly/io/async/ScopedEventBaseThread.h>
 #include <folly/io/async/test/MockAsyncTransport.h>
-#include <folly/ssl/Init.h>
 
 #include <quic/client/handshake/ClientTransportParametersExtension.h>
 #include <quic/client/state/ClientStateMachine.h>
@@ -48,8 +47,7 @@ class ClientHandshakeTest : public Test, public boost::static_visitor<> {
   ClientHandshakeTest() {}
 
   virtual void setupClientAndServerContext() {
-    clientCtx = std::make_shared<fizz::client::FizzClientContext>();
-    clientCtx->setClock(std::make_shared<fizz::test::MockClock>());
+    clientCtx = createClientCtx();
   }
 
   QuicVersion getVersion() {
@@ -75,7 +73,6 @@ class ClientHandshakeTest : public Test, public boost::static_visitor<> {
   }
 
   void SetUp() override {
-    folly::ssl::init();
     dg.reset(new DelayedHolder());
     serverCtx = ::quic::test::createServerCtx();
     serverCtx->setECHDecrypter(getECHDecrypter());
@@ -91,6 +88,7 @@ class ClientHandshakeTest : public Test, public boost::static_visitor<> {
                                 .setCertificateVerifier(verifier)
                                 .setPskCache(getPskCache())
                                 .setECHPolicy(getECHPolicy())
+                                .setECHRetryCallback(getECHRetryCallback())
                                 .build();
     conn.reset(new QuicClientConnectionState(handshakeFactory));
     conn->readCodec = std::make_unique<QuicReadCodec>(QuicNodeType::Client);
@@ -136,6 +134,11 @@ class ClientHandshakeTest : public Test, public boost::static_visitor<> {
     return nullptr;
   }
 
+  virtual std::shared_ptr<fizz::client::test::MockECHRetryCallback>
+  getECHRetryCallback() {
+    return nullptr;
+  }
+
   void clientServerRound() {
     auto writableBytes = getHandshakeWriteBytes();
     serverReadBuf.append(std::move(writableBytes));
@@ -298,6 +301,27 @@ class ClientHandshakeTest : public Test, public boost::static_visitor<> {
   std::shared_ptr<fizz::server::FizzServerContext> serverCtx;
 };
 
+TEST_F(ClientHandshakeTest, TestGetExportedKeyingMaterial) {
+  // Sanity check. getExportedKeyingMaterial () should return nullptr prior to
+  // an handshake.
+  auto ekm = handshake->getExportedKeyingMaterial(
+      "EXPORTER-Some-Label", folly::none, 32);
+  EXPECT_TRUE(!ekm.has_value());
+
+  clientServerRound();
+  serverClientRound();
+  handshake->handshakeConfirmed();
+  ekm = handshake->getExportedKeyingMaterial(
+      "EXPORTER-Some-Label", folly::none, 32);
+  ASSERT_TRUE(ekm.has_value());
+  EXPECT_EQ(ekm->size(), 32);
+
+  ekm = handshake->getExportedKeyingMaterial(
+      "EXPORTER-Some-Label", folly::ByteRange(), 32);
+  ASSERT_TRUE(ekm.has_value());
+  EXPECT_EQ(ekm->size(), 32);
+}
+
 TEST_F(ClientHandshakeTest, TestHandshakeSuccess) {
   EXPECT_CALL(*verifier, verify(_));
 
@@ -340,11 +364,25 @@ TEST_F(ClientHandshakeTest, TestRetryIntegrityVerification) {
   LongHeader header(
       LongHeader::Types::Retry, scid, dcid, 0, version, retryToken);
 
-  std::string integrityTag =
-      "\xd1\x69\x26\xd8\x1f\x6f\x9c\xa2\x95\x3a\x8a\xa4\x57\x5e\x1e\x49";
+  RetryPacket::IntegrityTagType integrityTag = {
+      0xd1,
+      0x69,
+      0x26,
+      0xd8,
+      0x1f,
+      0x6f,
+      0x9c,
+      0xa2,
+      0x95,
+      0x3a,
+      0x8a,
+      0xa4,
+      0x57,
+      0x5e,
+      0x1e,
+      0x49};
 
-  RetryPacket retryPacket(
-      std::move(header), folly::IOBuf::copyBuffer(integrityTag), initialByte);
+  RetryPacket retryPacket(std::move(header), integrityTag, initialByte);
 
   std::vector<uint8_t> odcidVec = {
       0x83, 0x94, 0xc8, 0xf0, 0x3e, 0x51, 0x57, 0x08};
@@ -392,10 +430,9 @@ TEST_F(ClientHandshakeTest, TestAppBytesInterpretedAsHandshake) {
 class ClientHandshakeCallbackTest : public ClientHandshakeTest {
  public:
   void setupClientAndServerContext() override {
-    clientCtx = std::make_shared<fizz::client::FizzClientContext>();
+    clientCtx = createClientCtx();
     clientCtx->setSupportedVersions({fizz::ProtocolVersion::tls_1_3});
     serverCtx->setSupportedVersions({fizz::ProtocolVersion::tls_1_3});
-    clientCtx->setClock(std::make_shared<fizz::test::MockClock>());
     setupZeroRttOnServerCtx(*serverCtx, psk_);
   }
 
@@ -441,15 +478,13 @@ class ClientHandshakeHRRTest : public ClientHandshakeTest {
   ~ClientHandshakeHRRTest() override = default;
 
   void setupClientAndServerContext() override {
-    clientCtx = std::make_shared<fizz::client::FizzClientContext>();
+    clientCtx = createClientCtx();
     clientCtx->setSupportedGroups(
         {fizz::NamedGroup::secp256r1, fizz::NamedGroup::x25519});
     clientCtx->setDefaultShares({fizz::NamedGroup::secp256r1});
-    clientCtx->setClock(std::make_shared<fizz::test::MockClock>());
-    serverCtx = std::make_shared<fizz::server::FizzServerContext>();
+    serverCtx = createServerCtx();
     serverCtx->setFactory(std::make_shared<QuicFizzFactory>());
     serverCtx->setSupportedGroups({fizz::NamedGroup::x25519});
-    serverCtx->setClock(std::make_shared<fizz::test::MockClock>());
     setupCtxWithTestCert(*serverCtx);
   }
 };
@@ -493,22 +528,22 @@ class ClientHandshakeZeroRttTest : public ClientHandshakeTest {
   ~ClientHandshakeZeroRttTest() override = default;
 
   void setupClientAndServerContext() override {
-    clientCtx = std::make_shared<fizz::client::FizzClientContext>();
+    clientCtx = createClientCtx();
     clientCtx->setSupportedVersions({fizz::ProtocolVersion::tls_1_3});
     clientCtx->setSupportedAlpns({"h3", "hq"});
-    clientCtx->setClock(std::make_shared<fizz::test::MockClock>());
     serverCtx->setSupportedVersions({fizz::ProtocolVersion::tls_1_3});
     serverCtx->setSupportedAlpns({"h3"});
-    serverCtx->setClock(std::make_shared<fizz::test::MockClock>());
     setupCtxWithTestCert(*serverCtx);
     psk = setupZeroRttOnClientCtx(*clientCtx, hostname);
     setupZeroRttServer();
   }
 
   std::shared_ptr<QuicPskCache> getPskCache() override {
-    auto pskCache = std::make_shared<BasicQuicPskCache>();
-    pskCache->putPsk(hostname, psk);
-    return pskCache;
+    if (!pskCache_) {
+      pskCache_ = std::make_shared<BasicQuicPskCache>();
+      pskCache_->putPsk(hostname, psk);
+    }
+    return pskCache_;
   }
 
   void connect() override {
@@ -534,6 +569,7 @@ class ClientHandshakeZeroRttTest : public ClientHandshakeTest {
   }
 
   QuicCachedPsk psk;
+  std::shared_ptr<QuicPskCache> pskCache_;
 };
 
 TEST_F(ClientHandshakeZeroRttTest, TestZeroRttSuccess) {
@@ -590,11 +626,20 @@ class ClientHandshakeZeroRttRejectFail : public ClientHandshakeZeroRttTest {
 };
 
 TEST_F(ClientHandshakeZeroRttRejectFail, TestZeroRttRejectionParamsDontMatch) {
+  // Before the handshake, we have not check the early params.
+  ASSERT_FALSE(handshake->getCanResendZeroRtt().has_value());
   clientServerRound();
+  // The server hasn't rejected zero-rtt yet so we should still have the psk.
+  ASSERT_TRUE(pskCache_->getPsk(hostname).has_value());
+
   EXPECT_EQ(handshake->getPhase(), ClientHandshake::Phase::Initial);
   expectHandshakeCipher(false);
   expectZeroRttCipher(true, false);
-  EXPECT_THROW(serverClientRound(), QuicInternalException);
+  EXPECT_NO_THROW(serverClientRound());
+  // After the handshake with rtt rejection, we should have checked the early
+  // params and marked them as invalid
+  ASSERT_TRUE(handshake->getCanResendZeroRtt().has_value());
+  EXPECT_FALSE(handshake->getCanResendZeroRtt().value());
 }
 
 class ClientHandshakeECHPolicyTest : public ClientHandshakeCallbackTest {
@@ -616,14 +661,20 @@ class ClientHandshakeECHPolicyTest : public ClientHandshakeCallbackTest {
     return echPolicy;
   }
 
+  std::shared_ptr<fizz::client::test::MockECHRetryCallback>
+  getECHRetryCallback() override {
+    return echCallback;
+  }
+
   fizz::ech::ECHConfigContentDraft getECHConfigContent() {
     fizz::ech::HpkeSymmetricCipherSuite suite{
         fizz::hpke::KDFId::Sha256, fizz::hpke::AeadId::TLS_AES_128_GCM_SHA256};
     fizz::ech::ECHConfigContentDraft echConfigContent;
     echConfigContent.key_config.config_id = 0xFB;
     echConfigContent.key_config.kem_id = fizz::hpke::KEMId::secp256r1;
-    echConfigContent.key_config.public_key = fizz::detail::encodeECPublicKey(
-        ::fizz::test::getPublicKey(::fizz::test::kP256PublicKey));
+    echConfigContent.key_config.public_key =
+        fizz::openssl::detail::encodeECPublicKey(
+            ::fizz::test::getPublicKey(::fizz::test::kP256PublicKey));
     echConfigContent.key_config.cipher_suites = {suite};
     echConfigContent.maximum_name_length = 100;
     echConfigContent.public_name = folly::IOBuf::copyBuffer("public.dummy.com");
@@ -638,15 +689,18 @@ class ClientHandshakeECHPolicyTest : public ClientHandshakeCallbackTest {
   }
 
   std::shared_ptr<fizz::client::test::MockECHPolicy> echPolicy;
+  std::shared_ptr<fizz::client::test::MockECHRetryCallback> echCallback;
   std::shared_ptr<fizz::ech::ECHConfigManager> echDecrypter;
 };
 
 TEST_F(ClientHandshakeECHPolicyTest, TestECHPolicyHandshake) {
   echPolicy = std::make_shared<fizz::client::test::MockECHPolicy>();
+  echCallback = std::make_shared<fizz::client::test::MockECHRetryCallback>();
   EXPECT_CALL(*echPolicy, getConfig(_))
       .WillOnce(Return(std::vector<fizz::ech::ECHConfig>{getECHConfig()}));
 
-  auto kex = std::make_unique<fizz::OpenSSLECKeyExchange<fizz::P256>>();
+  auto kex = fizz::openssl::makeOpenSSLECKeyExchange<fizz::P256>();
+
   kex->setPrivateKey(fizz::test::getPrivateKey(fizz::test::kP256Key));
   echDecrypter = std::make_shared<fizz::ech::ECHConfigManager>();
   echDecrypter->addDecryptionConfig(
