@@ -173,7 +173,7 @@ class FakeOneRttHandshakeLayer : public FizzClientHandshake {
       transportParams = quicCachedPsk->transportParams;
     }
 
-    getFizzState().sni() = hostname;
+    const_cast<fizz::client::State&>(getState()).sni() = hostname;
 
     connected_ = true;
     writeDataToQuicStream(
@@ -227,11 +227,18 @@ class FakeOneRttHandshakeLayer : public FizzClientHandshake {
 
   void setOneRttWriteCipher(std::unique_ptr<Aead> oneRttWriteCipher) {
     oneRttWriteCipher_ = std::move(oneRttWriteCipher);
+    writeTrafficSecret_ = folly::IOBuf::copyBuffer(getRandSecret());
   }
 
   void setOneRttReadCipher(std::unique_ptr<Aead> oneRttReadCipher) {
     getClientConn()->readCodec->setOneRttReadCipher(
         std::move(oneRttReadCipher));
+    readTrafficSecret_ = folly::IOBuf::copyBuffer(getRandSecret());
+  }
+
+  void setNextOneRttReadCipher(std::unique_ptr<Aead> oneRttReadCipher) {
+    getClientConn()->readCodec->setNextOneRttReadCipher(
+        std::move(oneRttReadCipher));
   }
 
   void setHandshakeReadCipher(std::unique_ptr<Aead> handshakeReadCipher) {
@@ -276,8 +283,9 @@ class FakeOneRttHandshakeLayer : public FizzClientHandshake {
         std::move(oneRttReadHeaderCipher));
   }
 
-  void setZeroRttRejected(bool rejected) {
+  void setZeroRttRejected(bool rejected, bool canResendZeroRtt) {
     setZeroRttRejectedForTest(rejected);
+    setCanResendZeroRttForTest(canResendZeroRtt);
     if (rejected) {
       createServerTransportParameters();
     }
@@ -313,6 +321,10 @@ class FakeOneRttHandshakeLayer : public FizzClientHandshake {
     return params_;
   }
 
+  Buf getNextTrafficSecret(folly::ByteRange /*secret*/) const override {
+    return folly::IOBuf::copyBuffer(getRandSecret());
+  }
+
   void triggerOnNewCachedPsk() {
     fizz::client::NewCachedPsk psk;
     onNewCachedPsk(psk);
@@ -355,9 +367,12 @@ class FakeOneRttHandshakeLayer : public FizzClientHandshake {
   bool matchEarlyParameters() override {
     throw std::runtime_error("matchEarlyParameters not implemented");
   }
-  std::pair<std::unique_ptr<Aead>, std::unique_ptr<PacketNumberCipher>>
-  buildCiphers(CipherKind, folly::ByteRange) override {
-    throw std::runtime_error("buildCiphers not implemented");
+  std::unique_ptr<Aead> buildAead(CipherKind, folly::ByteRange) override {
+    return createNoOpAead();
+  }
+  std::unique_ptr<PacketNumberCipher> buildHeaderCipher(
+      folly::ByteRange) override {
+    throw std::runtime_error("buildHeaderCipher not implemented");
   }
 };
 
@@ -370,12 +385,22 @@ class QuicClientTransportTestBase : public virtual testing::Test {
   virtual ~QuicClientTransportTestBase() = default;
 
   struct TestReadData {
-    std::unique_ptr<folly::IOBuf> data;
+    ReceivedUdpPacket udpPacket;
     folly::SocketAddress addr;
     folly::Optional<int> err;
 
     TestReadData(folly::ByteRange dataIn, folly::SocketAddress addrIn)
-        : data(folly::IOBuf::copyBuffer(dataIn)), addr(std::move(addrIn)) {}
+        : udpPacket(folly::IOBuf::copyBuffer(dataIn)),
+          addr(std::move(addrIn)) {}
+
+    TestReadData(
+        const ReceivedUdpPacket&& udpPacketIn,
+        folly::SocketAddress addrIn)
+        : udpPacket(
+              udpPacketIn.buf.clone(),
+              udpPacketIn.timings,
+              udpPacketIn.tosValue),
+          addr(std::move(addrIn)) {}
 
     explicit TestReadData(int errIn) : err(errIn) {}
   };
@@ -427,8 +452,8 @@ class QuicClientTransportTestBase : public virtual testing::Test {
             errno = *socketReads[0].err;
             return -1;
           }
-          auto testData = std::move(socketReads[0].data);
-          testData->coalesce();
+          auto& udpPacket = socketReads[0].udpPacket;
+          auto testData = std::move(udpPacket.buf.front());
           size_t testDataLen = testData->length();
           memcpy(
               msg->msg_iov[0].iov_base, testData->data(), testData->length());
@@ -437,9 +462,18 @@ class QuicClientTransportTestBase : public virtual testing::Test {
                 static_cast<sockaddr_storage*>(msg->msg_name));
             msg->msg_namelen = msg_len;
           }
+#ifdef FOLLY_HAVE_MSG_ERRQUEUE
+          // Populate ToS control message as IPv6 TClass
+          auto cmsg = CMSG_FIRSTHDR(msg);
+          cmsg->cmsg_level = SOL_IPV6;
+          cmsg->cmsg_type = IPV6_TCLASS;
+          memcpy(
+              CMSG_DATA(cmsg), &udpPacket.tosValue, sizeof(udpPacket.tosValue));
+#endif
           socketReads.pop_front();
           return testDataLen;
         }));
+    ON_CALL(*sock, getRecvTos()).WillByDefault(testing::Return(true));
     EXPECT_EQ(client->getConn().selfConnectionIds.size(), 1);
     EXPECT_EQ(
         client->getConn().selfConnectionIds[0].connId,
@@ -514,12 +548,14 @@ class QuicClientTransportTestBase : public virtual testing::Test {
 
   virtual void setFakeHandshakeCiphers() {
     auto readAead = test::createNoOpAead();
+    auto nextReadAead = test::createNoOpAead();
     auto writeAead = test::createNoOpAead();
     auto handshakeReadAead = test::createNoOpAead();
     auto handshakeWriteAead = test::createNoOpAead();
     mockClientHandshake->setHandshakeReadCipher(std::move(handshakeReadAead));
     mockClientHandshake->setHandshakeWriteCipher(std::move(handshakeWriteAead));
     mockClientHandshake->setOneRttReadCipher(std::move(readAead));
+    mockClientHandshake->setNextOneRttReadCipher(std::move(nextReadAead));
     mockClientHandshake->setOneRttWriteCipher(std::move(writeAead));
 
     mockClientHandshake->setHandshakeReadHeaderCipher(
@@ -721,7 +757,9 @@ class QuicClientTransportTestBase : public virtual testing::Test {
       folly::SocketAddress* peer = nullptr) {
     for (const auto& packet : data.getPackets()) {
       deliverDataWithoutErrorCheck(
-          peer == nullptr ? serverAddr : *peer, packet.buf->coalesce(), writes);
+          peer == nullptr ? serverAddr : *peer,
+          packet.buf.clone()->coalesce(),
+          writes);
     }
   }
 
@@ -751,13 +789,44 @@ class QuicClientTransportTestBase : public virtual testing::Test {
     deliverData(peer == nullptr ? serverAddr : *peer, data, writes);
   }
 
+  void deliverDataWithoutErrorCheck(
+      const folly::SocketAddress& addr,
+      const ReceivedUdpPacket&& udpPacket,
+      bool writes = true) {
+    ASSERT_TRUE(networkReadCallback);
+    socketReads.emplace_back(std::move(udpPacket), addr);
+    networkReadCallback->onNotifyDataAvailable(*sock);
+    if (writes) {
+      loopForWrites();
+    }
+  }
+
+  void deliverData(
+      const folly::SocketAddress& addr,
+      const ReceivedUdpPacket&& udpPacket,
+      bool writes = true) {
+    deliverDataWithoutErrorCheck(addr, std::move(udpPacket), writes);
+    if (client->getConn().localConnectionError) {
+      bool idleTimeout = false;
+      const LocalErrorCode* localError =
+          client->getConn().localConnectionError->code.asLocalErrorCode();
+      if (localError) {
+        idleTimeout = (*localError == LocalErrorCode::IDLE_TIMEOUT);
+      }
+      if (!idleTimeout) {
+        throw std::runtime_error(
+            toString(client->getConn().localConnectionError->code));
+      }
+    }
+  }
+
   void deliverData(
       NetworkData&& data,
       bool writes = true,
       folly::SocketAddress* peer = nullptr) {
     for (const auto& packet : data.getPackets()) {
       deliverData(
-          peer == nullptr ? serverAddr : *peer, packet.buf->coalesce(), writes);
+          peer == nullptr ? serverAddr : *peer, std::move(packet), writes);
     }
   }
 
