@@ -220,7 +220,11 @@ folly::dynamic ReadAckFrameLog::toDynamic() const {
   }
   d["acked_ranges"] = ackRangeDynamic;
   d["frame_type"] = toQlogString(frameType);
-  if (frameType == FrameType::ACK_RECEIVE_TIMESTAMPS) {
+  if (frameType == FrameType::ACK_ECN) {
+    d["ecn_ect0"] = ecnECT0Count;
+    d["ecn_ect1"] = ecnECT1Count;
+    d["ecn_ce"] = ecnCECount;
+  } else if (frameType == FrameType::ACK_RECEIVE_TIMESTAMPS) {
     if (maybeLatestRecvdPacketTime.has_value()) {
       d["latest_recvd_packet_time"] =
           maybeLatestRecvdPacketTime.value().count();
@@ -243,6 +247,7 @@ folly::dynamic ReadAckFrameLog::toDynamic() const {
     d["timestamp_ranges"] = recvdPacketsTimestampRangesDynamic;
   }
   d["ack_delay"] = ackDelay.count();
+
   return d;
 }
 
@@ -256,7 +261,11 @@ folly::dynamic WriteAckFrameLog::toDynamic() const {
   }
   d["acked_ranges"] = ackRangeDynamic;
   d["frame_type"] = toQlogString(frameType);
-  if (frameType == FrameType::ACK_RECEIVE_TIMESTAMPS) {
+  if (frameType == FrameType::ACK_ECN) {
+    d["ecn_ect0"] = ecnECT0Count;
+    d["ecn_ect1"] = ecnECT1Count;
+    d["ecn_ce"] = ecnCECount;
+  } else if (frameType == FrameType::ACK_RECEIVE_TIMESTAMPS) {
     if (maybeLatestRecvdPacketTime.has_value()) {
       d["latest_recvd_packet_time"] =
           maybeLatestRecvdPacketTime.value().count();
@@ -521,7 +530,7 @@ QLogAppLimitedUpdateEvent::QLogAppLimitedUpdateEvent(
 folly::dynamic QLogAppLimitedUpdateEvent::toDynamic() const {
   folly::dynamic d = folly::dynamic::array(
       folly::to<std::string>(refTime.count()),
-      "APP_LIMITED_UPDATE",
+      "app_limited_update",
       toString(eventType));
   folly::dynamic data = folly::dynamic::object();
   data["app_limited"] = limited ? kAppLimited : kAppUnlimited;
@@ -541,7 +550,7 @@ QLogBandwidthEstUpdateEvent::QLogBandwidthEstUpdateEvent(
 folly::dynamic QLogBandwidthEstUpdateEvent::toDynamic() const {
   folly::dynamic d = folly::dynamic::array(
       folly::to<std::string>(refTime.count()),
-      "BANDIWDTH_EST_UPDATE",
+      "bandwidth_est_update",
       toString(eventType));
   folly::dynamic data = folly::dynamic::object();
   data["bandwidth_bytes"] = bytes;
@@ -948,6 +957,31 @@ folly::dynamic QLogPriorityUpdateEvent::toDynamic() const {
   return d;
 }
 
+QLogL4sWeightUpdateEvent::QLogL4sWeightUpdateEvent(
+    double l4sWeightIn,
+    uint32_t newECT1EchoedIn,
+    uint32_t newCEEchoedIn,
+    std::chrono::microseconds refTimeIn)
+    : l4sWeight_(l4sWeightIn),
+      newECT1Echoed_(newECT1EchoedIn),
+      newCEEchoed_(newCEEchoedIn) {
+  eventType = QLogEventType::L4sWeightUpdate;
+  refTime = refTimeIn;
+}
+
+folly::dynamic QLogL4sWeightUpdateEvent::toDynamic() const {
+  folly::dynamic d = folly::dynamic::array(
+      folly::to<std::string>(refTime.count()),
+      "metric_update",
+      toString(eventType));
+  folly::dynamic data = folly::dynamic::object();
+  data["weight"] = l4sWeight_;
+  data["new_ect1"] = newECT1Echoed_;
+  data["new_ce"] = newCEEchoed_;
+  d.push_back(std::move(data));
+  return d;
+}
+
 folly::StringPiece toString(QLogEventType type) {
   switch (type) {
     case QLogEventType::PacketSent:
@@ -994,6 +1028,8 @@ folly::StringPiece toString(QLogEventType type) {
       return "path_validation";
     case QLogEventType::PriorityUpdate:
       return "priority";
+    case QLogEventType::L4sWeightUpdate:
+      return "l4s_weight_update";
   }
   folly::assume_unreachable();
 }
