@@ -7,8 +7,7 @@
 
 #pragma once
 
-#include <folly/Portability.h>
-#include <folly/dynamic.h>
+#include <folly/json/dynamic.h>
 #include <quic/codec/Types.h>
 #include <quic/logging/QLoggerConstants.h>
 #include <memory>
@@ -198,6 +197,10 @@ class ReadAckFrameLog : public QLogFrame {
   folly::Optional<std::chrono::microseconds> maybeLatestRecvdPacketTime;
   folly::Optional<PacketNum> maybeLatestRecvdPacketNum;
   RecvdPacketsTimestampsRangeVec recvdPacketsTimestampRanges;
+  uint32_t ecnECT0Count;
+  uint32_t ecnECT1Count;
+  uint32_t ecnCECount;
+
   ReadAckFrameLog(
       const ReadAckFrame::Vec& ackBlocksIn,
       std::chrono::microseconds ackDelayIn,
@@ -205,13 +208,19 @@ class ReadAckFrameLog : public QLogFrame {
       folly::Optional<std::chrono::microseconds> maybeLatestRecvdPacketTimeIn =
           folly::none,
       folly::Optional<PacketNum> maybeLatestRecvdPacketNumIn = folly::none,
-      RecvdPacketsTimestampsRangeVec recvdPacketsTimestampRangesIn = {})
+      RecvdPacketsTimestampsRangeVec recvdPacketsTimestampRangesIn = {},
+      uint32_t ecnECT0CountIn = 0,
+      uint32_t ecnECT1CountIn = 0,
+      uint32_t ecnCECountIn = 0)
       : ackBlocks{ackBlocksIn},
         ackDelay{ackDelayIn},
         frameType{frameTypeIn},
         maybeLatestRecvdPacketTime{maybeLatestRecvdPacketTimeIn},
         maybeLatestRecvdPacketNum{maybeLatestRecvdPacketNumIn},
-        recvdPacketsTimestampRanges(recvdPacketsTimestampRangesIn) {}
+        recvdPacketsTimestampRanges(std::move(recvdPacketsTimestampRangesIn)),
+        ecnECT0Count(ecnECT0CountIn),
+        ecnECT1Count(ecnECT1CountIn),
+        ecnCECount(ecnCECountIn) {}
   ~ReadAckFrameLog() override = default;
   FOLLY_NODISCARD folly::dynamic toDynamic() const override;
 };
@@ -224,6 +233,9 @@ class WriteAckFrameLog : public QLogFrame {
   folly::Optional<std::chrono::microseconds> maybeLatestRecvdPacketTime;
   folly::Optional<PacketNum> maybeLatestRecvdPacketNum;
   RecvdPacketsTimestampsRangeVec recvdPacketsTimestampRanges;
+  uint32_t ecnECT0Count;
+  uint32_t ecnECT1Count;
+  uint32_t ecnCECount;
 
   WriteAckFrameLog(
       const WriteAckFrame::AckBlockVec& ackBlocksIn,
@@ -232,14 +244,19 @@ class WriteAckFrameLog : public QLogFrame {
       folly::Optional<std::chrono::microseconds> maybeLatestRecvdPacketTimeIn =
           folly::none,
       folly::Optional<PacketNum> maybeLatestRecvdPacketNumIn = folly::none,
-      RecvdPacketsTimestampsRangeVec recvdPacketsTimestampRangesIn = {})
+      RecvdPacketsTimestampsRangeVec recvdPacketsTimestampRangesIn = {},
+      uint32_t ecnECT0CountIn = 0,
+      uint32_t ecnECT1CountIn = 0,
+      uint32_t ecnCECountIn = 0)
       : ackBlocks{ackBlocksIn},
         ackDelay{ackDelayIn},
         frameType{frameTypeIn},
-
         maybeLatestRecvdPacketTime{maybeLatestRecvdPacketTimeIn},
         maybeLatestRecvdPacketNum{maybeLatestRecvdPacketNumIn},
-        recvdPacketsTimestampRanges{recvdPacketsTimestampRangesIn} {}
+        recvdPacketsTimestampRanges{std::move(recvdPacketsTimestampRangesIn)},
+        ecnECT0Count(ecnECT0CountIn),
+        ecnECT1Count(ecnECT1CountIn),
+        ecnCECount(ecnCECountIn) {}
   ~WriteAckFrameLog() override = default;
   FOLLY_NODISCARD folly::dynamic toDynamic() const override;
 };
@@ -379,7 +396,8 @@ enum class QLogEventType : uint32_t {
   BandwidthEstUpdate,
   ConnectionMigration,
   PathValidation,
-  PriorityUpdate
+  PriorityUpdate,
+  L4sWeightUpdate
 };
 
 folly::StringPiece toString(QLogEventType type);
@@ -746,4 +764,20 @@ class QLogPriorityUpdateEvent : public QLogEvent {
   bool incremental_;
 };
 
+class QLogL4sWeightUpdateEvent : public QLogEvent {
+ public:
+  explicit QLogL4sWeightUpdateEvent(
+      double l4sWeightIn,
+      uint32_t newECT1EchoedIn,
+      uint32_t newCEEchoedIn,
+      std::chrono::microseconds refTimeIn);
+  ~QLogL4sWeightUpdateEvent() override = default;
+
+  FOLLY_NODISCARD folly::dynamic toDynamic() const override;
+
+  double l4sWeight_;
+  uint32_t newECT1Echoed_;
+  uint32_t newCEEchoed_;
+};
+
 } // namespace quic
