@@ -39,6 +39,7 @@ class SocketObserverInterface {
     streamEvents = 9,
     acksProcessedEvents = 10,
     packetsReceivedEvents = 11,
+    l4sWeightUpdatedEvents = 12,
   };
   virtual ~SocketObserverInterface() = default;
 
@@ -232,15 +233,20 @@ class SocketObserverInterface {
       // Number of bytes in the received packet.
       const uint64_t packetNumBytes;
 
+      // TOS value
+      const uint8_t packetTos;
+
       struct BuilderFields {
         folly::Optional<TimePoint> maybePacketReceiveTime;
         folly::Optional<uint64_t> maybePacketNumBytes;
+        folly::Optional<uint8_t> maybePacketTos;
         explicit BuilderFields() = default;
       };
 
       struct Builder : public BuilderFields {
         Builder&& setPacketReceiveTime(const TimePoint packetReceiveTimeIn);
         Builder&& setPacketNumBytes(const uint64_t packetNumBytesIn);
+        Builder&& setPacketTos(const uint8_t packettos);
         ReceivedUdpPacket build() &&;
         explicit Builder() = default;
       };
@@ -413,6 +419,19 @@ class SocketObserverInterface {
   using StreamOpenEvent = StreamEvent;
   using StreamCloseEvent = StreamEvent;
 
+  struct L4sWeightUpdateEvent {
+    explicit L4sWeightUpdateEvent(
+        double l4sWeightIn,
+        uint32_t newECT1EchoedIn,
+        uint32_t newCEEchoedIn)
+        : l4sWeight(l4sWeightIn),
+          newECT1Echoed(newECT1EchoedIn),
+          newCEEchoed(newCEEchoedIn) {}
+    double l4sWeight;
+    uint32_t newECT1Echoed;
+    uint32_t newCEEchoed;
+  };
+
   /**
    * Events.
    */
@@ -580,6 +599,17 @@ class SocketObserverInterface {
   virtual void streamClosed(
       QuicSocket*, /* socket */
       const StreamCloseEvent& /* event */) {}
+
+  /**
+   * l4sWeightUpdated() is invoked when l4s in enabled and new CE markings are
+   * echoed by the peer
+   *
+   * @param socket   Socket associated with the event
+   * @param event    Event with the new weight and received ECN marks
+   */
+  virtual void l4sWeightUpdated(
+      QuicSocket* /* socket */,
+      const L4sWeightUpdateEvent& /* event */) noexcept {}
 };
 
 } // namespace quic
