@@ -6,6 +6,7 @@
  */
 
 #include <quic/congestion_control/Bbr.h>
+#include <quic/congestion_control/ServerCongestionControllerFactory.h>
 #include <quic/dsr/frontend/WriteFunctions.h>
 #include <quic/fizz/server/handshake/FizzServerQuicHandshakeContext.h>
 #include <quic/server/QuicServerTransport.h>
@@ -24,8 +25,8 @@
 namespace quic {
 
 QuicServerTransport::QuicServerTransport(
-    std::shared_ptr<FollyQuicEventBase> evb,
-    std::unique_ptr<FollyQuicAsyncUDPSocket> sock,
+    std::shared_ptr<QuicEventBase> evb,
+    std::unique_ptr<QuicAsyncUDPSocket> sock,
     folly::MaybeManagedPtr<ConnectionSetupCallback> connSetupCb,
     folly::MaybeManagedPtr<ConnectionCallback> connStreamsCb,
     std::shared_ptr<const fizz::server::FizzServerContext> ctx,
@@ -42,8 +43,8 @@ QuicServerTransport::QuicServerTransport(
 }
 
 QuicServerTransport::QuicServerTransport(
-    std::shared_ptr<FollyQuicEventBase> evb,
-    std::unique_ptr<FollyQuicAsyncUDPSocket> sock,
+    std::shared_ptr<QuicEventBase> evb,
+    std::unique_ptr<QuicAsyncUDPSocket> sock,
     folly::MaybeManagedPtr<ConnectionSetupCallback> connSetupCb,
     folly::MaybeManagedPtr<ConnectionCallback> connStreamsCb,
     std::shared_ptr<const fizz::server::FizzServerContext> ctx,
@@ -126,6 +127,9 @@ void QuicServerTransport::setTransportStatsCallback(
     QuicTransportStatsCallback* statsCallback) noexcept {
   if (conn_) {
     conn_->statsCallback = statsCallback;
+    if (conn_->readCodec) {
+      conn_->readCodec->setConnectionStatsCallback(statsCallback);
+    }
   }
 }
 
@@ -190,6 +194,7 @@ void QuicServerTransport::onReadData(
   maybeNotifyConnectionIdRetired();
   maybeIssueConnectionIds();
   maybeNotifyTransportReady();
+  maybeUpdateCongestionControllerFromTicket();
 }
 
 void QuicServerTransport::accept() {
@@ -267,63 +272,21 @@ void QuicServerTransport::writeData() {
   // use.
   SCOPE_EXIT {
     conn_->pendingEvents.numProbePackets = {};
+    maybeInitiateKeyUpdate(*conn_);
   };
   if (conn_->initialWriteCipher) {
-    auto& initialCryptoStream =
-        *getCryptoStream(*conn_->cryptoState, EncryptionLevel::Initial);
-    CryptoStreamScheduler initialScheduler(*conn_, initialCryptoStream);
-    auto& numProbePackets =
-        conn_->pendingEvents.numProbePackets[PacketNumberSpace::Initial];
-    if ((numProbePackets && initialCryptoStream.retransmissionBuffer.size() &&
-         conn_->outstandings.packetCount[PacketNumberSpace::Initial]) ||
-        initialScheduler.hasData() || toWriteInitialAcks(*conn_)) {
-      CHECK(conn_->initialWriteCipher);
-      CHECK(conn_->initialHeaderCipher);
-
-      auto res = writeCryptoAndAckDataToSocket(
-          *socket_,
-          *conn_,
-          srcConnId /* src */,
-          destConnId /* dst */,
-          LongHeader::Types::Initial,
-          *conn_->initialWriteCipher,
-          *conn_->initialHeaderCipher,
-          version,
-          packetLimit);
-
-      packetLimit -= res.packetsWritten;
-      serverConn_->numHandshakeBytesSent += res.bytesWritten;
-    }
+    auto res = handleInitialWriteDataCommon(srcConnId, destConnId, packetLimit);
+    packetLimit -= res.packetsWritten;
+    serverConn_->numHandshakeBytesSent += res.bytesWritten;
     if (!packetLimit && !conn_->pendingEvents.anyProbePackets()) {
       return;
     }
   }
   if (conn_->handshakeWriteCipher) {
-    auto& handshakeCryptoStream =
-        *getCryptoStream(*conn_->cryptoState, EncryptionLevel::Handshake);
-    CryptoStreamScheduler handshakeScheduler(*conn_, handshakeCryptoStream);
-    auto& numProbePackets =
-        conn_->pendingEvents.numProbePackets[PacketNumberSpace::Handshake];
-    if ((conn_->outstandings.packetCount[PacketNumberSpace::Handshake] &&
-         handshakeCryptoStream.retransmissionBuffer.size() &&
-         numProbePackets) ||
-        handshakeScheduler.hasData() || toWriteHandshakeAcks(*conn_)) {
-      CHECK(conn_->handshakeWriteCipher);
-      CHECK(conn_->handshakeWriteHeaderCipher);
-      auto res = writeCryptoAndAckDataToSocket(
-          *socket_,
-          *conn_,
-          srcConnId /* src */,
-          destConnId /* dst */,
-          LongHeader::Types::Handshake,
-          *conn_->handshakeWriteCipher,
-          *conn_->handshakeWriteHeaderCipher,
-          version,
-          packetLimit);
-
-      packetLimit -= res.packetsWritten;
-      serverConn_->numHandshakeBytesSent += res.bytesWritten;
-    }
+    auto res =
+        handleHandshakeWriteDataCommon(srcConnId, destConnId, packetLimit);
+    packetLimit -= res.packetsWritten;
+    serverConn_->numHandshakeBytesSent += res.bytesWritten;
     if (!packetLimit && !conn_->pendingEvents.anyProbePackets()) {
       return;
     }
@@ -443,7 +406,7 @@ void QuicServerTransport::onCryptoEventAvailable() noexcept {
       VLOG(10) << "Got crypto event after connection closed " << *this;
       return;
     }
-    FOLLY_MAYBE_UNUSED auto self = sharedGuard();
+    [[maybe_unused]] auto self = sharedGuard();
     updateHandshakeState(*serverConn_);
     processPendingData(false);
     // pending data may contain connection close
@@ -514,9 +477,7 @@ void QuicServerTransport::processPendingData(bool async) {
       for (auto& pendingPacket : *pendingData) {
         serverPtr->onNetworkData(
             pendingPacket.peer,
-            NetworkData(
-                std::move(pendingPacket.udpPacket.buf),
-                pendingPacket.udpPacket.timings.receiveTimePoint));
+            NetworkData(std::move(pendingPacket.udpPacket)));
         if (serverPtr->closeState_ == CloseState::CLOSED) {
           // The pending data could potentially contain a connection close, or
           // the app could have triggered a connection close with an error. It
@@ -690,6 +651,43 @@ void QuicServerTransport::maybeNotifyTransportReady() {
   }
 }
 
+namespace {
+uint64_t determineCwndFromHint(const TransportSettings& tp, uint64_t cwndHint) {
+  if (cwndHint <= tp.cwndModerateJumpstart) {
+    // Weak Connectivity History
+    return tp.cwndWeakJumpstart;
+  } else if (cwndHint <= tp.cwndStrongJumpstart) {
+    // Moderate Connectivity History
+    return tp.cwndModerateJumpstart;
+  } else {
+    // Strong Connectivity History
+    return tp.cwndStrongJumpstart;
+  }
+}
+} // namespace
+
+void QuicServerTransport::maybeUpdateCongestionControllerFromTicket() {
+  if (serverConn_->transportSettings.useCwndHintsInSessionTicket &&
+      serverConn_->maybeCwndHintBytes.has_value() &&
+      serverConn_->lossState.maybeLrtt.has_value()) {
+    auto newCwnd = determineCwndFromHint(
+        serverConn_->transportSettings,
+        serverConn_->maybeCwndHintBytes.value());
+    serverConn_->maybeCwndHintBytes.reset();
+    if (newCwnd > conn_->congestionController->getCongestionWindow()) {
+      conn_->transportSettings.initCwndInMss =
+          newCwnd / conn_->udpSendPacketLen;
+      if (conn_->pacer) {
+        conn_->pacer.get()->refreshPacingRate(
+            newCwnd, serverConn_->lossState.lrtt);
+      }
+      conn_->congestionController =
+          conn_->congestionControllerFactory.get()->makeCongestionController(
+              *conn_, conn_->congestionController.get()->type());
+    }
+  }
+}
+
 void QuicServerTransport::registerTransportKnobParamHandler(
     uint64_t paramId,
     std::function<void(QuicServerTransport*, TransportKnobParam::Val)>&&
@@ -711,6 +709,15 @@ QuicServerTransport::getPeerCertificate() const {
   return nullptr;
 }
 
+const std::shared_ptr<const folly::AsyncTransportCertificate>
+QuicServerTransport::getSelfCertificate() const {
+  const auto handshakeLayer = serverConn_->serverHandshakeLayer;
+  if (handshakeLayer) {
+    return handshakeLayer->getState().serverCert();
+  }
+  return nullptr;
+}
+
 void QuicServerTransport::onTransportKnobs(Buf knobBlob) {
   if (knobBlob->length() > 0) {
     std::string serializedKnobs = std::string(
@@ -1098,6 +1105,48 @@ void QuicServerTransport::registerAllTransportKnobParamHandlers() {
         VLOG(3) << "CONNECTION_MIGRATION KnobParam received: "
                 << static_cast<bool>(val);
       });
+  registerTransportKnobParamHandler(
+      static_cast<uint64_t>(TransportKnobParamId::KEY_UPDATE_INTERVAL),
+      [](QuicServerTransport* serverTransport, TransportKnobParam::Val value) {
+        CHECK(serverTransport);
+        auto val = std::get<uint64_t>(value);
+        if (val < 1000 || val > 8ul * 1000 * 1000) {
+          std::string errMsg = fmt::format(
+              "KEY_UPDATE_INTERVAL KnobParam received with invalid value: {}",
+              val);
+          throw std::runtime_error(errMsg);
+        }
+        auto server_conn = serverTransport->serverConn_;
+        server_conn->transportSettings.initiateKeyUpdate = val > 0;
+        server_conn->transportSettings.keyUpdatePacketCountInterval = val;
+        VLOG(3) << "KEY_UPDATE_INTERVAL KnobParam received: " << val;
+      });
+  registerTransportKnobParamHandler(
+      static_cast<uint64_t>(
+          TransportKnobParamId::USE_NEW_STREAM_BLOCKED_CONDITION),
+      [](QuicServerTransport* serverTransport, TransportKnobParam::Val value) {
+        CHECK(serverTransport);
+        bool useNewStreamBlockedCondition =
+            static_cast<bool>(std::get<uint64_t>(value));
+        auto server_conn = serverTransport->serverConn_;
+        server_conn->transportSettings.useNewStreamBlockedCondition =
+            useNewStreamBlockedCondition;
+        VLOG(3) << "USE_NEW_STREAM_BLOCKED_CONDITION KnobParam received: "
+                << useNewStreamBlockedCondition;
+      });
+  registerTransportKnobParamHandler(
+      static_cast<uint64_t>(
+          TransportKnobParamId::AUTOTUNE_RECV_STREAM_FLOW_CONTROL),
+      [](QuicServerTransport* serverTransport, TransportKnobParam::Val value) {
+        CHECK(serverTransport);
+        bool autotuneReceiveStreamFlowControl =
+            static_cast<bool>(std::get<uint64_t>(value));
+        auto server_conn = serverTransport->serverConn_;
+        server_conn->transportSettings.autotuneReceiveStreamFlowControl =
+            autotuneReceiveStreamFlowControl;
+        VLOG(3) << "AUTOTUNE_RECV_STREAM_FLOW_CONTROL KnobParam received: "
+                << autotuneReceiveStreamFlowControl;
+      });
 }
 
 QuicConnectionStats QuicServerTransport::getConnectionsStats() const {
@@ -1146,4 +1195,29 @@ void QuicServerTransport::logTimeBasedStats() const {
   }
 }
 
+folly::Optional<std::vector<TransportParameter>>
+QuicServerTransport::getPeerTransportParams() const {
+  if (serverConn_ && serverConn_->serverHandshakeLayer) {
+    auto maybeParams =
+        serverConn_->serverHandshakeLayer->getClientTransportParams();
+    if (maybeParams) {
+      return maybeParams->parameters;
+    }
+  }
+  return folly::none;
+}
+
+void QuicServerTransport::setCongestionControl(CongestionControlType type) {
+  if (!conn_->congestionControllerFactory) {
+    // If you are hitting this, update your application to call
+    // setCongestionControllerFactory() on the transport to share one factory
+    // for all transports.
+    conn_->congestionControllerFactory =
+        std::make_shared<ServerCongestionControllerFactory>();
+    LOG(WARNING)
+        << "A congestion controller factory is not set. Using a default per-transport instance.";
+  }
+  QuicTransportBase::setCongestionControl(type);
+}
+
 } // namespace quic
