@@ -13,6 +13,7 @@
 #include <folly/system/ThreadId.h>
 #include <quic/QuicConstants.h>
 #include <atomic>
+#include <chrono>
 #include <memory>
 
 #ifdef FOLLY_HAVE_MSG_ERRQUEUE
@@ -89,7 +90,7 @@ void QuicServerWorker::bind(
     case SetEventCallback::RECVMSG_MULTISHOT:
       socket_->setRecvmsgMultishotCallback(this);
       break;
-  };
+  }
   // TODO this totally doesn't work, we can't apply socket options before
   // bind, since bind creates the fd.
   if (socketOptions_) {
@@ -99,6 +100,9 @@ void QuicServerWorker::bind(
         address.getFamily(),
         folly::SocketOptionKey::ApplyPos::PRE_BIND);
   }
+  if (transportSettings_.readEcnOnIngress) {
+    socket_->setRecvTos(true);
+  }
   socket_->bind(address, bindOptions);
   if (socketOptions_) {
     applySocketOptions(
@@ -306,14 +310,19 @@ void QuicServerWorker::onDataAvailable(
   auto originalPacketReceiveTime = packetReceiveTime;
   if (params.ts) {
     // This is the software system time from the datagram.
-    auto packetNowDuration =
+    auto packetRxEpochUs =
         folly::to<std::chrono::microseconds>(params.ts.value()[0]);
-    auto wallNowDuration =
-        std::chrono::duration_cast<std::chrono::microseconds>(
-            std::chrono::system_clock::now().time_since_epoch());
-    auto durationSincePacketNow = wallNowDuration - packetNowDuration;
-    if (packetNowDuration != 0us && durationSincePacketNow > 0us) {
-      packetReceiveTime -= durationSincePacketNow;
+    if (packetRxEpochUs != 0us) {
+      auto now = std::chrono::system_clock::now();
+      auto nowEpochUs = std::chrono::duration_cast<std::chrono::microseconds>(
+          now.time_since_epoch());
+      auto rxDelayUs = nowEpochUs - packetRxEpochUs;
+      if (rxDelayUs >= 0us) {
+        packetReceiveTime -= rxDelayUs;
+        QUIC_STATS(statsCallback_, onRxDelaySample, rxDelayUs.count());
+      } else {
+        VLOG(10) << "Negative rx delay: " << rxDelayUs.count() << "us";
+      }
     }
   }
   // System time can move backwards, so we want to make sure that the receive
@@ -341,7 +350,10 @@ void QuicServerWorker::onDataAvailable(
     data->append(len);
     QUIC_STATS(statsCallback_, onPacketReceived);
     QUIC_STATS(statsCallback_, onRead, len);
-    handleNetworkData(client, std::move(data), packetReceiveTime);
+    ReceivedUdpPacket udpPacket(std::move(data));
+    udpPacket.timings.receiveTimePoint = packetReceiveTime;
+    udpPacket.tosValue = params.tos;
+    handleNetworkData(client, udpPacket);
   } else {
     // if we receive a truncated packet
     // we still need to consider the prev valid ones
@@ -350,7 +362,6 @@ void QuicServerWorker::onDataAvailable(
     if (truncated) {
       len -= len % params.gro;
     }
-
     data->append(len);
     QUIC_STATS(statsCallback_, onPacketReceived);
     QUIC_STATS(statsCallback_, onRead, len);
@@ -363,7 +374,10 @@ void QuicServerWorker::onDataAvailable(
         // start at offset, use all the remaining data
         data->trimStart(offset);
         DCHECK_EQ(data->length(), remaining);
-        handleNetworkData(client, std::move(data), packetReceiveTime);
+        ReceivedUdpPacket udpPacket(std::move(data));
+        udpPacket.timings.receiveTimePoint = packetReceiveTime;
+        udpPacket.tosValue = params.tos;
+        handleNetworkData(client, udpPacket);
         break;
       }
       auto tmp = data->cloneOne();
@@ -375,15 +389,17 @@ void QuicServerWorker::onDataAvailable(
       DCHECK_EQ(tmp->length(), params.gro);
       offset += params.gro;
       remaining -= params.gro;
-      handleNetworkData(client, std::move(tmp), packetReceiveTime);
+      ReceivedUdpPacket udpPacket(std::move(tmp));
+      udpPacket.timings.receiveTimePoint = packetReceiveTime;
+      udpPacket.tosValue = params.tos;
+      handleNetworkData(client, udpPacket);
     }
   }
 }
 
 void QuicServerWorker::handleNetworkData(
     const folly::SocketAddress& client,
-    Buf data,
-    const TimePoint& packetReceiveTime,
+    ReceivedUdpPacket& udpPacket,
     bool isForwardedData) noexcept {
   // if packet drop reason is set, invoke stats cb accordingly
   auto packetDropReason = PacketDropReason::NONE;
@@ -395,7 +411,7 @@ void QuicServerWorker::handleNetworkData(
 
   try {
     // check error conditions for packet drop & early return
-    folly::io::Cursor cursor(data.get());
+    folly::io::Cursor cursor(udpPacket.buf.front());
     if (shutdown_) {
       VLOG(4) << "Packet received after shutdown, dropping";
       packetDropReason = PacketDropReason::SERVER_SHUTDOWN;
@@ -430,7 +446,7 @@ void QuicServerWorker::handleNetworkData(
         return forwardNetworkData(
             client,
             std::move(routingData),
-            NetworkData(std::move(data), packetReceiveTime),
+            NetworkData(std::move(udpPacket)),
             folly::none, /* quicVersion */
             isForwardedData);
       }
@@ -449,7 +465,7 @@ void QuicServerWorker::handleNetworkData(
       }
 
       if (maybeSendVersionNegotiationPacketOrDrop(
-              client, isInitial, invariant, data->computeChainDataLength())) {
+              client, isInitial, invariant, udpPacket.buf.chainLength())) {
         return;
       }
 
@@ -470,12 +486,12 @@ void QuicServerWorker::handleNetworkData(
       return forwardNetworkData(
           client,
           std::move(routingData),
-          NetworkData(std::move(data), packetReceiveTime),
+          NetworkData(std::move(udpPacket)),
           invariant.version,
           isForwardedData);
     }
 
-    if (!tryHandlingAsHealthCheck(client, *data)) {
+    if (!tryHandlingAsHealthCheck(client, *udpPacket.buf.front())) {
       VLOG(6) << "Failed to parse long header";
       packetDropReason = PacketDropReason::PARSE_ERROR_LONG_HEADER;
     }
@@ -650,8 +666,10 @@ QuicServerTransport::Ptr QuicServerWorker::makeTransport(
     }
     trans->setCongestionControllerFactory(ccFactory_);
     trans->setTransportStatsCallback(statsCallback_.get()); // ok if nullptr
-    if (quicVersion == QuicVersion::MVFST_EXPERIMENTAL) {
-      transportSettings_.initCwndInMss = 45;
+    if (quicVersion == QuicVersion::MVFST_EXPERIMENTAL3) {
+      // Use twice the default pacing gain to make BBRv2's startup behavior
+      // similar to BBRv1's.
+      transportSettings_.ccaConfig.overrideStartupPacingGain = 2 * 2.89;
     }
 
     auto transportSettings = transportSettingsOverrideFn_
@@ -666,7 +684,7 @@ QuicServerTransport::Ptr QuicServerWorker::makeTransport(
         << (transportSettings.dataPathType == DataPathType::ContinuousMemory
                 ? "ContinuousMemory"
                 : "ChainedMemory");
-    if (quicVersion == QuicVersion::MVFST_EXPERIMENTAL2) {
+    if (quicVersion == QuicVersion::MVFST_EXPERIMENTAL) {
       transportSettings.includeCwndHintsInSessionTicket = true;
       transportSettings.useCwndHintsInSessionTicket = true;
     }
@@ -841,7 +859,7 @@ void QuicServerWorker::dispatchPacketData(
   // route with destinationConnId chosen by the peer and IP address of the
   // peer.
   CHECK(routingData.headerForm == HeaderForm::Long);
-  auto sit = sourceAddressMap_.find({client, dstConnId});
+  auto sit = sourceAddressMap_.find(std::make_pair(client, dstConnId));
   if (sit != sourceAddressMap_.end()) {
     VLOG(4) << "Found existing connection for client=" << client << " "
             << sit->second.get();
@@ -892,7 +910,7 @@ void QuicServerWorker::dispatchPacketData(
 
   // If there is a token present, decrypt it (could be either a retry
   // token or a new token)
-  folly::io::Cursor cursor(networkData.getPackets().front().buf.get());
+  folly::io::Cursor cursor(networkData.getPackets().front().buf.front());
   auto maybeEncryptedToken = maybeGetEncryptedToken(cursor);
   bool hasTokenSecret = transportSettings_.retryTokenSecret.hasValue();
 
@@ -952,10 +970,15 @@ void QuicServerWorker::sendResetPacket(
     return;
   }
   auto packetSize = networkData.getTotalData();
+  if (packetSize <= kMinStatelessPacketSize) {
+    // We must decrease the packet size to prevent reset loops. If it's already
+    // too small, we can't send one.
+    return;
+  }
   auto resetSize = std::min<uint16_t>(packetSize, kDefaultMaxUDPPayload);
   // Per the spec, less than 43 we should respond with packet size - 1.
   if (packetSize < 43) {
-    resetSize = std::max<uint16_t>(packetSize - 1, kMinStatelessPacketSize);
+    resetSize = packetSize - 1;
   } else {
     resetSize = std::max<uint16_t>(
         folly::Random::secureRand32() % resetSize, kMinStatelessPacketSize);
@@ -1087,8 +1110,12 @@ void QuicServerWorker::sendRetryPacket(
       folly::IOBuf::copyBuffer(encryptedTokenStr));
   Buf pseudoRetryPacketBuf = std::move(pseudoBuilder).buildPacket();
   FizzRetryIntegrityTagGenerator fizzRetryIntegrityTagGenerator;
-  auto integrityTag = fizzRetryIntegrityTagGenerator.getRetryIntegrityTag(
+  auto integrityTagBuf = fizzRetryIntegrityTagGenerator.getRetryIntegrityTag(
       QuicVersion::MVFST_INVALID, pseudoRetryPacketBuf.get());
+  folly::io::Cursor cursor{integrityTagBuf.get()};
+
+  RetryPacket::IntegrityTagType integrityTag = {0};
+  cursor.pull(integrityTag.data(), integrityTag.size());
 
   // Create the actual retry packet
   RetryPacketBuilder builder(
@@ -1096,7 +1123,7 @@ void QuicServerWorker::sendRetryPacket(
       srcConnId, /* dst conn id */
       QuicVersion::MVFST_INVALID,
       std::move(encryptedTokenStr),
-      std::move(integrityTag));
+      integrityTag);
 
   auto retryData = std::move(builder).buildPacket();
   auto retryDataLen = retryData->computeChainDataLength();
