@@ -20,7 +20,6 @@
 #include <folly/io/async/SSLContext.h>
 #include <folly/io/async/ScopedEventBaseThread.h>
 #include <folly/io/async/test/MockAsyncTransport.h>
-#include <folly/ssl/Init.h>
 
 #include <quic/QuicConstants.h>
 #include <quic/common/test/TestUtils.h>
@@ -74,11 +73,11 @@ class ServerHandshakeTest : public Test {
   }
 
   void SetUp() override {
-    folly::ssl::init();
-    clientCtx = std::make_shared<fizz::client::FizzClientContext>();
+    // This client context is used outside the context of QUIC in this test, so
+    // we have to manually configure the QUIC record customizations.
+    clientCtx = quic::test::createClientCtx();
     clientCtx->setOmitEarlyRecordLayer(true);
     clientCtx->setFactory(std::make_shared<QuicFizzFactory>());
-    clientCtx->setClock(std::make_shared<fizz::test::MockClock>());
     serverCtx = quic::test::createServerCtx();
     setupClientAndServerContext();
     auto fizzServerContext = FizzServerQuicHandshakeContext::Builder()
@@ -152,18 +151,16 @@ class ServerHandshakeTest : public Test {
     try {
       setHandshakeState();
       waitForData = false;
-      auto writableBytes = getHandshakeWriteBytes();
-      while (writableBytes && !writableBytes->empty() && !waitForData) {
+      do {
+        auto writableBytes = getHandshakeWriteBytes();
+        if (writableBytes->empty()) {
+          break;
+        }
         VLOG(1) << "server->client bytes="
                 << writableBytes->computeChainDataLength();
         clientReadBuffer.append(std::move(writableBytes));
-        if (!clientReadBuffer.empty()) {
-          fizzClient->newTransportData();
-        }
-        if (!waitForData) {
-          writableBytes = getHandshakeWriteBytes();
-        }
-      }
+        fizzClient->newTransportData();
+      } while (!waitForData);
     } catch (const QuicTransportException& e) {
       VLOG(1) << "server exception " << e.what();
       ex = std::make_exception_ptr(e);
@@ -202,24 +199,22 @@ class ServerHandshakeTest : public Test {
     inRoundScope_ = true;
     evb.loop();
     waitForData = false;
-    auto writableBytes = getHandshakeWriteBytes();
-    while (writableBytes && !writableBytes->empty() && !waitForData) {
+    do {
+      auto writableBytes = getHandshakeWriteBytes();
+      if (writableBytes->empty()) {
+        break;
+      }
       VLOG(1) << "server->client bytes="
               << writableBytes->computeChainDataLength();
       clientReadBuffer.append(std::move(writableBytes));
-      if (!clientReadBuffer.empty()) {
-        fizzClient->newTransportData();
-      }
-      if (!waitForData) {
-        writableBytes = getHandshakeWriteBytes();
-      }
-    }
+      fizzClient->newTransportData();
+    } while (!waitForData);
     evb.loop();
   }
 
   void setHandshakeState() {
-    auto oneRttWriteCipherTmp = handshake->getOneRttWriteCipher();
-    auto oneRttReadCipherTmp = handshake->getOneRttReadCipher();
+    auto oneRttWriteCipherTmp = handshake->getFirstOneRttWriteCipher();
+    auto oneRttReadCipherTmp = handshake->getFirstOneRttReadCipher();
     auto zeroRttReadCipherTmp = handshake->getZeroRttReadCipher();
     auto handshakeWriteCipherTmp = std::move(conn->handshakeWriteCipher);
     auto handshakeReadCipherTmp = handshake->getHandshakeReadCipher();
@@ -327,6 +322,8 @@ class ServerHandshakeTest : public Test {
 
   void operator()(fizz::EndOfData&) {}
 
+  void operator()(fizz::client::ECHRetryAvailable&) {}
+
   class DelayedHolder : public folly::DelayedDestruction {};
 
   std::unique_ptr<DelayedHolder, folly::DelayedDestruction::Destructor> dg;
@@ -370,6 +367,26 @@ class ServerHandshakeTest : public Test {
   bool waitForData{false};
 };
 
+TEST_F(ServerHandshakeTest, TestGetExportedKeyingMaterial) {
+  // Sanity check. getExportedKeyingMaterial() should return nullptr prior to
+  // an handshake.
+  auto ekm = handshake->getExportedKeyingMaterial(
+      "EXPORTER-Some-Label", folly::none, 32);
+  EXPECT_TRUE(!ekm.has_value());
+
+  clientServerRound();
+  serverClientRound();
+  ekm = handshake->getExportedKeyingMaterial(
+      "EXPORTER-Some-Label", folly::none, 32);
+  ASSERT_TRUE(ekm.has_value());
+  EXPECT_EQ(ekm->size(), 32);
+
+  ekm = handshake->getExportedKeyingMaterial(
+      "EXPORTER-Some-Label", folly::ByteRange(), 32);
+  ASSERT_TRUE(ekm.has_value());
+  EXPECT_EQ(ekm->size(), 32);
+}
+
 TEST_F(ServerHandshakeTest, TestHandshakeSuccess) {
   clientServerRound();
   EXPECT_EQ(handshake->getPhase(), ServerHandshake::Phase::Handshake);
@@ -380,7 +397,7 @@ TEST_F(ServerHandshakeTest, TestHandshakeSuccess) {
     std::rethrow_exception(ex);
   }
   expectOneRttCipher(true);
-  EXPECT_EQ(handshake->getApplicationProtocol(), folly::none);
+  EXPECT_EQ(handshake->getApplicationProtocol(), "quic_test");
   EXPECT_TRUE(handshakeSuccess);
 }
 
@@ -401,7 +418,7 @@ TEST_F(ServerHandshakeTest, TestHandshakeSuccessIgnoreNonHandshake) {
     std::rethrow_exception(ex);
   }
   expectOneRttCipher(true);
-  EXPECT_EQ(handshake->getApplicationProtocol(), folly::none);
+  EXPECT_EQ(handshake->getApplicationProtocol(), "quic_test");
   EXPECT_TRUE(handshakeSuccess);
 }
 
@@ -691,7 +708,7 @@ TEST_F(ServerHandshakeAsyncErrorTest, TestAsyncError) {
   EXPECT_CALL(serverCallback, onCryptoEventAvailable())
       .WillRepeatedly(Invoke([&] {
         try {
-          handshake->getOneRttReadCipher();
+          handshake->getFirstOneRttReadCipher();
         } catch (std::exception&) {
           error = true;
         }
@@ -712,7 +729,7 @@ TEST_F(ServerHandshakeAsyncErrorTest, TestCancelOnAsyncError) {
       }));
   promise.setValue();
   evb.loop();
-  EXPECT_THROW(handshake->getOneRttReadCipher(), std::runtime_error);
+  EXPECT_THROW(handshake->getFirstOneRttReadCipher(), std::runtime_error);
 }
 
 TEST_F(ServerHandshakeAsyncErrorTest, TestCancelWhileWaitingAsyncError) {
@@ -723,7 +740,7 @@ TEST_F(ServerHandshakeAsyncErrorTest, TestCancelWhileWaitingAsyncError) {
 
   promise.setValue();
   evb.loop();
-  EXPECT_THROW(handshake->getOneRttReadCipher(), std::runtime_error);
+  EXPECT_THROW(handshake->getFirstOneRttReadCipher(), std::runtime_error);
 }
 
 class ServerHandshakeSyncErrorTest : public ServerHandshakePskTest {
@@ -742,7 +759,7 @@ TEST_F(ServerHandshakeSyncErrorTest, TestError) {
   // Make an async ticket decryption operation.
   clientServerRound();
   evb.loop();
-  EXPECT_THROW(handshake->getOneRttReadCipher(), std::runtime_error);
+  EXPECT_THROW(handshake->getFirstOneRttReadCipher(), std::runtime_error);
 }
 
 class ServerHandshakeZeroRttDefaultAppTokenValidatorTest
