@@ -93,10 +93,12 @@ void recoverOrResetCongestionAndRttState(
 void maybeSetExperimentalSettings(QuicServerConnectionState& conn) {
   // no-op versions
   if (conn.version == QuicVersion::MVFST_EXPERIMENTAL) {
-    // MVFST_EXPERIMENTAL is used to set initCwnd
-    // QuicServerWorker.cpp before CC is initialized.
+    // MVFST_EXPERIMENTAL is used to jumpstart the cwnd for clients based on a
+    // cwndHint from a previous connection in QuicServerTransport.cpp
   } else if (conn.version == QuicVersion::MVFST_EXPERIMENTAL2) {
   } else if (conn.version == QuicVersion::MVFST_EXPERIMENTAL3) {
+    // MVFST_EXPERIMENTAL3 is used to apply a 2x pace scaling for BBRv2
+    // QuicServerWorker.cpp before CC is initialized.
   }
 }
 
@@ -233,7 +235,8 @@ void processClientInitialParams(
       TransportParameterId::receive_timestamps_exponent,
       clientParams.parameters);
   if (conn.version == QuicVersion::QUIC_V1 ||
-      conn.version == QuicVersion::QUIC_V1_ALIAS) {
+      conn.version == QuicVersion::QUIC_V1_ALIAS ||
+      conn.version == QuicVersion::QUIC_V1_ALIAS2) {
     auto initialSourceConnId = getConnIdParameter(
         TransportParameterId::initial_source_connection_id,
         clientParams.parameters);
@@ -339,16 +342,8 @@ void processClientInitialParams(
     maxUdpPayloadSize = std::min(*packetSize, maxUdpPayloadSize);
     conn.peerMaxUdpPayloadSize = maxUdpPayloadSize;
     if (conn.transportSettings.canIgnorePathMTU) {
-      if (*packetSize > kDefaultMaxUDPPayload) {
-        // A good peer should never set oversized limit, so to be safe we
-        // fallback to default
-        conn.udpSendPacketLen = kDefaultUDPSendPacketLen;
-      } else {
-        // Otherwise, canIgnorePathMTU forces us to immediately set
-        // udpSendPacketLen
-        // TODO: rename "canIgnorePathMTU" to "forciblySetPathMTU"
-        conn.udpSendPacketLen = maxUdpPayloadSize;
-      }
+      *packetSize = std::min<uint64_t>(*packetSize, kDefaultMaxUDPPayload);
+      conn.udpSendPacketLen = *packetSize;
     }
   }
 
@@ -386,9 +381,9 @@ void updateHandshakeState(QuicServerConnectionState& conn) {
   // However, the cipher is only exported to QUIC if early data attempt is
   // accepted. Otherwise, the cipher will be available after cfin is
   // processed.
-  auto oneRttWriteCipher = handshakeLayer->getOneRttWriteCipher();
+  auto oneRttWriteCipher = handshakeLayer->getFirstOneRttWriteCipher();
   // One RTT read cipher is available after cfin is processed.
-  auto oneRttReadCipher = handshakeLayer->getOneRttReadCipher();
+  auto oneRttReadCipher = handshakeLayer->getFirstOneRttReadCipher();
 
   auto oneRttWriteHeaderCipher = handshakeLayer->getOneRttWriteHeaderCipher();
   auto oneRttReadHeaderCipher = handshakeLayer->getOneRttReadHeaderCipher();
@@ -419,6 +414,7 @@ void updateHandshakeState(QuicServerConnectionState& conn) {
           "Duplicate 1-rtt write cipher", TransportErrorCode::CRYPTO_ERROR);
     }
     conn.oneRttWriteCipher = std::move(oneRttWriteCipher);
+    conn.oneRttWritePhase = ProtectionType::KeyPhaseZero;
 
     updatePacingOnKeyEstablished(conn);
 
@@ -440,6 +436,8 @@ void updateHandshakeState(QuicServerConnectionState& conn) {
     conn.isClientAddrVerified = true;
     conn.writableBytesLimit.reset();
     conn.readCodec->setOneRttReadCipher(std::move(oneRttReadCipher));
+    conn.readCodec->setNextOneRttReadCipher(
+        handshakeLayer->getNextOneRttReadCipher());
   }
   auto handshakeReadCipher = handshakeLayer->getHandshakeReadCipher();
   auto handshakeReadHeaderCipher =
@@ -586,7 +584,6 @@ void onConnectionMigration(
     throw QuicTransportException(
         "Too many migrations", TransportErrorCode::INVALID_MIGRATION);
   }
-  ++conn.migrationState.numMigrations;
 
   bool hasPendingPathChallenge = conn.pendingEvents.pathChallenge.has_value();
   // Clear any pending path challenge frame that is not sent
@@ -598,6 +595,7 @@ void onConnectionMigration(
       previousPeerAddresses.end(),
       newPeerAddress);
   if (it == previousPeerAddresses.end()) {
+    ++conn.migrationState.numMigrations;
     // send new path challenge
     conn.pendingEvents.pathChallenge.emplace(folly::Random::secureRand64());
 
@@ -710,7 +708,9 @@ static void handleCipherUnavailable(
     ServerEvents::ReadData pendingReadData;
     pendingReadData.peer = readData.peer;
     pendingReadData.udpPacket = ReceivedUdpPacket(
-        std::move(originalData->packet), readData.udpPacket.timings);
+        std::move(originalData->packet),
+        readData.udpPacket.timings,
+        readData.udpPacket.tosValue);
     pendingData->emplace_back(std::move(pendingReadData));
     VLOG(10) << "Adding pending data to "
              << toString(originalData->protectionType)
@@ -734,15 +734,14 @@ void onServerReadDataFromOpen(
     ServerEvents::ReadData& readData) {
   CHECK_EQ(conn.state, ServerState::Open);
   // Don't bother parsing if the data is empty.
-  if (!readData.udpPacket.buf ||
-      readData.udpPacket.buf->computeChainDataLength() == 0) {
+  if (readData.udpPacket.buf.empty()) {
     return;
   }
   bool firstPacketFromPeer = false;
   if (!conn.readCodec) {
     firstPacketFromPeer = true;
 
-    folly::io::Cursor cursor(readData.udpPacket.buf.get());
+    folly::io::Cursor cursor(readData.udpPacket.buf.front());
     auto initialByte = cursor.readBE<uint8_t>();
     auto parsedLongHeader = parseLongHeaderInvariant(initialByte, cursor);
     if (!parsedLongHeader) {
@@ -828,6 +827,7 @@ void onServerReadDataFromOpen(
     const CryptoFactory& cryptoFactory =
         conn.serverHandshakeLayer->getCryptoFactory();
     conn.readCodec = std::make_unique<QuicReadCodec>(QuicNodeType::Server);
+    conn.readCodec->setConnectionStatsCallback(conn.statsCallback);
     conn.readCodec->setInitialReadCipher(cryptoFactory.getClientInitialCipher(
         initialDestinationConnectionId, version));
     conn.readCodec->setClientConnectionId(clientConnectionId);
@@ -850,8 +850,8 @@ void onServerReadDataFromOpen(
         initialDestinationConnectionId, version);
     conn.peerAddress = conn.originalPeerAddress;
   }
-  BufQueue udpData;
-  udpData.append(std::move(readData.udpPacket.buf));
+  BufQueue& udpData = readData.udpPacket.buf;
+  uint64_t processedPacketsTotal = 0;
   for (uint16_t processedPackets = 0;
        !udpData.empty() && processedPackets < kMaxNumCoalescedPackets;
        processedPackets++) {
@@ -1016,8 +1016,8 @@ void onServerReadDataFromOpen(
     }
 
     auto& ackState = getAckState(conn, packetNumberSpace);
-    uint64_t distanceFromExpectedPacketNum = addPacketToAckState(
-        conn, ackState, packetNum, readData.udpPacket.timings);
+    uint64_t distanceFromExpectedPacketNum =
+        addPacketToAckState(conn, ackState, packetNum, readData.udpPacket);
     if (distanceFromExpectedPacketNum > 0) {
       QUIC_STATS(conn.statsCallback, onOutOfOrderPacketReceived);
     }
@@ -1028,6 +1028,70 @@ void onServerReadDataFromOpen(
     bool isNonProbingPacket = false;
     bool handshakeConfirmedThisLoop = false;
 
+    AckedPacketVisitor ackedPacketVisitor =
+        [&](const OutstandingPacketWrapper& outstandingPacket) {
+          maybeVerifyPendingKeyUpdate(conn, outstandingPacket, regularPacket);
+        };
+
+    AckedFrameVisitor ackedFrameVisitor =
+        [&](const OutstandingPacketWrapper&,
+            const QuicWriteFrame& packetFrame) {
+          switch (packetFrame.type()) {
+            case QuicWriteFrame::Type::WriteStreamFrame: {
+              const WriteStreamFrame& frame = *packetFrame.asWriteStreamFrame();
+              VLOG(4) << "Server received ack for stream=" << frame.streamId
+                      << " offset=" << frame.offset << " fin=" << frame.fin
+                      << " len=" << frame.len << " " << conn;
+              auto ackedStream = conn.streamManager->getStream(frame.streamId);
+              if (ackedStream) {
+                sendAckSMHandler(*ackedStream, frame);
+              }
+              break;
+            }
+            case QuicWriteFrame::Type::WriteCryptoFrame: {
+              const WriteCryptoFrame& frame = *packetFrame.asWriteCryptoFrame();
+              auto cryptoStream =
+                  getCryptoStream(*conn.cryptoState, encryptionLevel);
+              processCryptoStreamAck(*cryptoStream, frame.offset, frame.len);
+              break;
+            }
+            case QuicWriteFrame::Type::RstStreamFrame: {
+              const RstStreamFrame& frame = *packetFrame.asRstStreamFrame();
+              VLOG(4) << "Server received ack for reset stream="
+                      << frame.streamId << " " << conn;
+              auto stream = conn.streamManager->getStream(frame.streamId);
+              if (stream) {
+                sendRstAckSMHandler(*stream);
+              }
+              break;
+            }
+            case QuicWriteFrame::Type::WriteAckFrame: {
+              const WriteAckFrame& frame = *packetFrame.asWriteAckFrame();
+              DCHECK(!frame.ackBlocks.empty());
+              VLOG(4) << "Server received ack for largestAcked="
+                      << frame.ackBlocks.front().end << " " << conn;
+              commonAckVisitorForAckFrame(ackState, frame);
+              break;
+            }
+            case QuicWriteFrame::Type::PingFrame:
+              conn.pendingEvents.cancelPingTimeout = true;
+              return;
+            case QuicWriteFrame::Type::QuicSimpleFrame: {
+              const QuicSimpleFrame& frame = *packetFrame.asQuicSimpleFrame();
+              // ACK of HandshakeDone is a server-specific behavior.
+              if (frame.asHandshakeDoneFrame()) {
+                // Call handshakeConfirmed outside of the packet
+                // processing loop to avoid a re-entrancy.
+                handshakeConfirmedThisLoop = true;
+              }
+              break;
+            }
+            default: {
+              break;
+            }
+          }
+        };
+
     for (auto& quicFrame : regularPacket.frames) {
       switch (quicFrame.type()) {
         case QuicFrame::Type::ReadAckFrame: {
@@ -1039,71 +1103,8 @@ void onServerReadDataFromOpen(
               conn,
               packetNumberSpace,
               ackFrame,
-              [&](const OutstandingPacketWrapper&,
-                  const QuicWriteFrame& packetFrame,
-                  const ReadAckFrame&) {
-                switch (packetFrame.type()) {
-                  case QuicWriteFrame::Type::WriteStreamFrame: {
-                    const WriteStreamFrame& frame =
-                        *packetFrame.asWriteStreamFrame();
-                    VLOG(4)
-                        << "Server received ack for stream=" << frame.streamId
-                        << " offset=" << frame.offset << " fin=" << frame.fin
-                        << " len=" << frame.len << " " << conn;
-                    auto ackedStream =
-                        conn.streamManager->getStream(frame.streamId);
-                    if (ackedStream) {
-                      sendAckSMHandler(*ackedStream, frame);
-                    }
-                    break;
-                  }
-                  case QuicWriteFrame::Type::WriteCryptoFrame: {
-                    const WriteCryptoFrame& frame =
-                        *packetFrame.asWriteCryptoFrame();
-                    auto cryptoStream =
-                        getCryptoStream(*conn.cryptoState, encryptionLevel);
-                    processCryptoStreamAck(
-                        *cryptoStream, frame.offset, frame.len);
-                    break;
-                  }
-                  case QuicWriteFrame::Type::RstStreamFrame: {
-                    const RstStreamFrame& frame =
-                        *packetFrame.asRstStreamFrame();
-                    VLOG(4) << "Server received ack for reset stream="
-                            << frame.streamId << " " << conn;
-                    auto stream = conn.streamManager->getStream(frame.streamId);
-                    if (stream) {
-                      sendRstAckSMHandler(*stream);
-                    }
-                    break;
-                  }
-                  case QuicWriteFrame::Type::WriteAckFrame: {
-                    const WriteAckFrame& frame = *packetFrame.asWriteAckFrame();
-                    DCHECK(!frame.ackBlocks.empty());
-                    VLOG(4) << "Server received ack for largestAcked="
-                            << frame.ackBlocks.front().end << " " << conn;
-                    commonAckVisitorForAckFrame(ackState, frame);
-                    break;
-                  }
-                  case QuicWriteFrame::Type::PingFrame:
-                    conn.pendingEvents.cancelPingTimeout = true;
-                    return;
-                  case QuicWriteFrame::Type::QuicSimpleFrame: {
-                    const QuicSimpleFrame& frame =
-                        *packetFrame.asQuicSimpleFrame();
-                    // ACK of HandshakeDone is a server-specific behavior.
-                    if (frame.asHandshakeDoneFrame()) {
-                      // Call handshakeConfirmed outside of the packet
-                      // processing loop to avoid a re-entrancy.
-                      handshakeConfirmedThisLoop = true;
-                    }
-                    break;
-                  }
-                  default: {
-                    break;
-                  }
-                }
-              },
+              ackedPacketVisitor,
+              ackedFrameVisitor,
               markPacketLoss,
               readData.udpPacket.timings.receiveTimePoint));
           break;
@@ -1258,8 +1259,8 @@ void onServerReadDataFromOpen(
         }
         case QuicFrame::Type::DatagramFrame: {
           DatagramFrame& frame = *quicFrame.asDatagramFrame();
-          VLOG(10) << "Server received datagram data: "
-                   << " len=" << frame.length;
+          VLOG(10) << "Server received datagram data: " << " len="
+                   << frame.length;
           // Datagram isn't retransmittable. But we would like to ack them
           // early. So, make Datagram frames count towards ack policy
           pktHasRetransmittableData = true;
@@ -1296,6 +1297,8 @@ void onServerReadDataFromOpen(
       handshakeConfirmed(conn);
     }
 
+    maybeHandleIncomingKeyUpdate(conn);
+
     // Update writable limit before processing the handshake data. This is so
     // that if we haven't decided whether or not to validate the peer, we won't
     // increase the limit.
@@ -1368,7 +1371,10 @@ void onServerReadDataFromOpen(
       conn.readCodec->setInitialHeaderCipher(nullptr);
       implicitAckCryptoStream(conn, EncryptionLevel::Initial);
     }
-    QUIC_STATS(conn.statsCallback, onPacketProcessed);
+    processedPacketsTotal++;
+  }
+  if (processedPacketsTotal > 0) {
+    QUIC_STATS(conn.statsCallback, onPacketsProcessed, processedPacketsTotal);
   }
   VLOG_IF(4, !udpData.empty())
       << "Leaving " << udpData.chainLength()
@@ -1380,8 +1386,7 @@ void onServerReadDataFromClosed(
     QuicServerConnectionState& conn,
     ServerEvents::ReadData& readData) {
   CHECK_EQ(conn.state, ServerState::Closed);
-  BufQueue udpData;
-  udpData.append(std::move(readData.udpPacket.buf));
+  BufQueue& udpData = readData.udpPacket.buf;
   auto packetSize = udpData.empty() ? 0 : udpData.chainLength();
   if (!conn.readCodec) {
     // drop data. We closed before we even got the first packet. This is
@@ -1493,6 +1498,8 @@ void onServerReadDataFromClosed(
     conn.qLogger->addPacket(regularPacket, packetSize);
   }
 
+  // TODO: Should we honor a key update from the peer on a closed connection?
+
   // Only process the close frames in the packet
   for (auto& quicFrame : regularPacket.frames) {
     switch (quicFrame.type()) {
