@@ -22,15 +22,82 @@ namespace {
  */
 struct OutstandingPacketWithHandlerContext {
   explicit OutstandingPacketWithHandlerContext(
-      OutstandingPacketWrapper outstandingPacketIn)
-      : outstandingPacket(std::move(outstandingPacketIn)) {}
+      OutstandingPacketWrapper* outstandingPacketIn)
+      : outstandingPacket(outstandingPacketIn) {}
 
-  OutstandingPacketWrapper outstandingPacket;
+  OutstandingPacketWrapper* outstandingPacket;
   bool processAllFrames{false};
 };
 
 } // namespace
 
+void removeOutstandingsForAck(
+    QuicConnectionStateBase& conn,
+    PacketNumberSpace pnSpace,
+    const ReadAckFrame& frame) {
+  auto currentPacketIt = getLastOutstandingPacket(
+      conn,
+      pnSpace,
+      true /* includeDeclaredLost */,
+      true /* includeScheduledForDestruction */);
+
+  auto ackBlockIt = frame.ackBlocks.cbegin();
+  while (ackBlockIt != frame.ackBlocks.cend() &&
+         currentPacketIt != conn.outstandings.packets.rend()) {
+    // In reverse order, find the first outstanding packet that has a packet
+    // number LE the endPacket of the current ack range.
+    auto rPacketIt = std::lower_bound(
+        currentPacketIt,
+        conn.outstandings.packets.rend(),
+        ackBlockIt->endPacket,
+        [&](const auto& packetWithTime, const auto& val) {
+          return packetWithTime.packet.header.getPacketSequenceNum() > val;
+        });
+    if (rPacketIt == conn.outstandings.packets.rend()) {
+      return;
+    }
+
+    auto eraseEnd = rPacketIt;
+    while (rPacketIt != conn.outstandings.packets.rend()) {
+      auto currentPacketNum = rPacketIt->packet.header.getPacketSequenceNum();
+      auto currentPacketNumberSpace =
+          rPacketIt->packet.header.getPacketNumberSpace();
+      if (pnSpace != currentPacketNumberSpace) {
+        // When the next packet is not in the same packet number space, we need
+        // to skip it in current ack processing. If the iterator has moved, that
+        // means we have found packets in the current space that are acked by
+        // this ack block. So the code erases the current iterator range and
+        // move the iterator to be the next search point.
+        if (rPacketIt != eraseEnd) {
+          auto nextElem = conn.outstandings.packets.erase(
+              rPacketIt.base(), eraseEnd.base());
+          rPacketIt = std::reverse_iterator<decltype(nextElem)>(nextElem) + 1;
+          eraseEnd = rPacketIt;
+        } else {
+          rPacketIt++;
+          eraseEnd = rPacketIt;
+        }
+        continue;
+      }
+      if (currentPacketNum < ackBlockIt->startPacket) {
+        break;
+      }
+      CHECK(rPacketIt->metadata.scheduledForDestruction);
+      CHECK_GT(conn.outstandings.scheduledForDestructionCount, 0);
+      conn.outstandings.scheduledForDestructionCount--;
+      rPacketIt++;
+    }
+    if (rPacketIt != eraseEnd) {
+      auto nextElem =
+          conn.outstandings.packets.erase(rPacketIt.base(), eraseEnd.base());
+      currentPacketIt = std::reverse_iterator<decltype(nextElem)>(nextElem);
+    } else {
+      currentPacketIt = rPacketIt;
+    }
+    ackBlockIt++;
+  }
+}
+
 /**
  * Process ack frame and acked outstanding packets.
  *
@@ -51,11 +118,14 @@ AckEvent processAckFrame(
     QuicConnectionStateBase& conn,
     PacketNumberSpace pnSpace,
     const ReadAckFrame& frame,
-    const AckVisitor& ackVisitor,
+    const AckedPacketVisitor& ackedPacketVisitor,
+    const AckedFrameVisitor& ackedFrameVisitor,
     const LossVisitor& lossVisitor,
     const TimePoint& ackReceiveTime) {
   const auto nowTime = Clock::now();
 
+  updateEcnCountEchoed(conn, pnSpace, frame);
+
   // TODO: send error if we get an ack for a packet we've not sent t18721184
   auto ack = AckEvent::Builder()
                  .setAckTime(ackReceiveTime)
@@ -64,12 +134,18 @@ AckEvent processAckFrame(
                  .setPacketNumberSpace(pnSpace)
                  .setLargestAckedPacket(frame.largestAcked)
                  .setIsImplicitAck(frame.implicit)
+                 .setEcnCounts(
+                     frame.ecnECT0Count, frame.ecnECT1Count, frame.ecnCECount)
                  .build();
 
   // temporary storage to enable packets to be processed in sent order
   SmallVec<OutstandingPacketWithHandlerContext, 50> packetsWithHandlerContext;
 
-  auto currentPacketIt = getLastOutstandingPacketIncludingLost(conn, pnSpace);
+  auto currentPacketIt = getLastOutstandingPacket(
+      conn,
+      pnSpace,
+      true /* includeDeclaredLost */,
+      true /* includeScheduledForDestruction */);
 
   // Store first outstanding packet number to ignore old receive timestamps.
   const auto& firstOutstandingPacket =
@@ -112,13 +188,11 @@ AckEvent processAckFrame(
       VLOG(10) << __func__ << " less than all outstanding packets outstanding="
                << conn.outstandings.numOutstanding() << " range=["
                << ackBlockIt->startPacket << ", " << ackBlockIt->endPacket
-               << "]"
-               << " " << conn;
+               << "]" << " " << conn;
       ackBlockIt++;
       break;
     }
 
-    auto eraseEnd = rPacketIt;
     while (rPacketIt != conn.outstandings.packets.rend()) {
       auto currentPacketNum = rPacketIt->packet.header.getPacketSequenceNum();
       auto currentPacketNumberSpace =
@@ -129,20 +203,14 @@ AckEvent processAckFrame(
         // means we have found packets in the current space that are acked by
         // this ack block. So the code erases the current iterator range and
         // move the iterator to be the next search point.
-        if (rPacketIt != eraseEnd) {
-          auto nextElem = conn.outstandings.packets.erase(
-              rPacketIt.base(), eraseEnd.base());
-          rPacketIt = std::reverse_iterator<decltype(nextElem)>(nextElem) + 1;
-          eraseEnd = rPacketIt;
-        } else {
-          rPacketIt++;
-          eraseEnd = rPacketIt;
-        }
+        rPacketIt++;
         continue;
       }
       if (currentPacketNum < ackBlockIt->startPacket) {
         break;
       }
+      rPacketIt->metadata.scheduledForDestruction = true;
+      conn.outstandings.scheduledForDestructionCount++;
       VLOG(10) << __func__ << " acked packetNum=" << currentPacketNum
                << " space=" << currentPacketNumberSpace << " handshake="
                << (int)((rPacketIt->metadata.isHandshake) ? 1 : 0) << " "
@@ -292,32 +360,23 @@ AckEvent processAckFrame(
       ack.totalBytesAcked = conn.lossState.totalBytesAcked;
 
       {
+        OutstandingPacketWrapper* wrapper = &(*rPacketIt);
         auto tmpIt = packetsWithHandlerContext.emplace(
             std::find_if(
                 packetsWithHandlerContext.rbegin(),
                 packetsWithHandlerContext.rend(),
                 [&currentPacketNum](const auto& packetWithHandlerContext) {
-                  return packetWithHandlerContext.outstandingPacket.packet
+                  return packetWithHandlerContext.outstandingPacket->packet
                              .header.getPacketSequenceNum() > currentPacketNum;
                 })
                 .base(),
-            std::move(*rPacketIt));
+            OutstandingPacketWithHandlerContext(wrapper));
         tmpIt->processAllFrames = needsProcess;
       }
 
       rPacketIt++;
     }
-    // Done searching for acked outstanding packets in current ack block. Erase
-    // the current iterator range which is the last batch of continuous
-    // outstanding packets that are in this ack block. Move the iterator to be
-    // the next search point.
-    if (rPacketIt != eraseEnd) {
-      auto nextElem =
-          conn.outstandings.packets.erase(rPacketIt.base(), eraseEnd.base());
-      currentPacketIt = std::reverse_iterator<decltype(nextElem)>(nextElem);
-    } else {
-      currentPacketIt = rPacketIt;
-    }
+    currentPacketIt = rPacketIt;
     ackBlockIt++;
   }
 
@@ -336,9 +395,16 @@ AckEvent processAckFrame(
        packetWithHandlerContextItr != packetsWithHandlerContext.rend();
        packetWithHandlerContextItr++) {
     auto& outstandingPacket = packetWithHandlerContextItr->outstandingPacket;
+
+    // run the ACKed packet visitor
+    ackedPacketVisitor(*outstandingPacket);
+
+    // Update ecn counts
+    incrementEcnCountForAckedPacket(conn, pnSpace);
+
     const auto processAllFrames = packetWithHandlerContextItr->processAllFrames;
     AckEvent::AckPacket::DetailsPerStream detailsPerStream;
-    for (auto& packetFrame : outstandingPacket.packet.frames) {
+    for (auto& packetFrame : outstandingPacket->packet.frames) {
       if (!processAllFrames &&
           packetFrame.type() != QuicWriteFrame::Type::WriteAckFrame) {
         continue; // skip processing this frame
@@ -388,51 +454,24 @@ AckEvent processAckFrame(
             getLargestDeliverableOffset(*maybeAckedStreamState)};
       }(packetFrame);
 
-      // run the ACK visitor
-      ackVisitor(outstandingPacket, packetFrame, frame);
+      // run the ACKed frame visitor
+      ackedFrameVisitor(*outstandingPacket, packetFrame);
 
       // Part 2 and 3: Process current state relative to the PreAckVistorState.
       if (maybePreAckVisitorState.has_value()) {
         const auto& preAckVisitorState = maybePreAckVisitorState.value();
         const WriteStreamFrame& ackedFrame = *packetFrame.asWriteStreamFrame();
 
-        // determine if this frame was a retransmission
-        const bool retransmission = ([&outstandingPacket, &ackedFrame]() {
-          // in some cases (some unit tests), stream details are not available
-          // in these cases, we assume it is not a retransmission
-          if (const auto* maybeStreamDetails = folly::get_ptr(
-                  outstandingPacket.metadata.detailsPerStream,
-                  ackedFrame.streamId)) {
-            const auto& maybeFirstNewStreamByteOffset =
-                maybeStreamDetails->maybeFirstNewStreamByteOffset;
-            return (
-                !maybeFirstNewStreamByteOffset.has_value() ||
-                maybeFirstNewStreamByteOffset.value() > ackedFrame.offset);
-          }
-          return false; // assume not a retransmission
-        })();
-
         // check for change in ACK IntervalSet version
         CHECK(maybeAckedStreamState);
         if (preAckVisitorState.ackIntervalSetVersion !=
             getAckIntervalSetVersion(*maybeAckedStreamState)) {
           // we were able to fill in a hole in the ACK interval
-          detailsPerStream.recordFrameDelivered(ackedFrame, retransmission);
-
-          // check for change in delivery offset
-          const auto maybeLargestDeliverableOffset =
-              getLargestDeliverableOffset(*maybeAckedStreamState);
-          if (preAckVisitorState.maybeLargestDeliverableOffset !=
-              maybeLargestDeliverableOffset) {
-            CHECK(maybeLargestDeliverableOffset.has_value());
-            detailsPerStream.recordDeliveryOffsetUpdate(
-                ackedFrame.streamId, maybeLargestDeliverableOffset.value());
-          }
+          detailsPerStream.recordFrameDelivered(ackedFrame);
         } else {
           // we got an ACK of a frame that was already marked as delivered
           // when handling the ACK of some earlier packet; mark as such
-          detailsPerStream.recordFrameAlreadyDelivered(
-              ackedFrame, retransmission);
+          detailsPerStream.recordFrameAlreadyDelivered(ackedFrame);
 
           // should be no change in delivery offset
           DCHECK(
@@ -442,18 +481,20 @@ AckEvent processAckFrame(
       }
     }
     auto maybeRxTimestamp = packetReceiveTimeStamps.find(
-        outstandingPacket.packet.header.getPacketSequenceNum());
+        outstandingPacket->packet.header.getPacketSequenceNum());
     ack.ackedPackets.emplace_back(
         CongestionController::AckEvent::AckPacket::Builder()
             .setPacketNum(
-                outstandingPacket.packet.header.getPacketSequenceNum())
+                outstandingPacket->packet.header.getPacketSequenceNum())
             .setNonDsrPacketSequenceNumber(
-                outstandingPacket.nonDsrPacketSequenceNumber.value_or(0))
-            .setOutstandingPacketMetadata(std::move(outstandingPacket.metadata))
+                outstandingPacket->nonDsrPacketSequenceNumber.value_or(0))
+            .setOutstandingPacketMetadata(outstandingPacket->metadata)
             .setDetailsPerStream(std::move(detailsPerStream))
             .setLastAckedPacketInfo(
-                std::move(outstandingPacket.lastAckedPacketInfo))
-            .setAppLimited(outstandingPacket.isAppLimited)
+                outstandingPacket->lastAckedPacketInfo
+                    ? &outstandingPacket->lastAckedPacketInfo.value()
+                    : nullptr)
+            .setAppLimited(outstandingPacket->isAppLimited)
             .setReceiveDeltaTimeStamp(
                 maybeRxTimestamp != packetReceiveTimeStamps.end()
                     ? folly::make_optional(
@@ -530,6 +571,8 @@ AckEvent processAckFrame(
     }
   }
 
+  removeOutstandingsForAck(conn, pnSpace, frame);
+
   return ack;
 }
 
@@ -548,7 +591,8 @@ void clearOldOutstandingPackets(
       if (time < opItr->metadata.time) {
         break;
       }
-      if (opItr->packet.header.getPacketNumberSpace() != pnSpace) {
+      if (opItr->packet.header.getPacketNumberSpace() != pnSpace ||
+          opItr->metadata.scheduledForDestruction) {
         if (eraseBegin != opItr) {
           // We want to keep [eraseBegin, opItr) within a single PN space.
           opItr = conn.outstandings.packets.erase(eraseBegin, opItr);
@@ -714,4 +758,30 @@ void commonAckVisitorForAckFrame(
     }
   }
 }
+
+void incrementEcnCountForAckedPacket(
+    QuicConnectionStateBase& conn,
+    PacketNumberSpace pnSpace) {
+  if (conn.ecnState == ECNState::NotAttempted ||
+      conn.ecnState == ECNState::FailedValidation) {
+    // Nothing to update.
+    return;
+  }
+  auto& ackState = getAckState(conn, pnSpace);
+  ackState.minimumExpectedEcnMarksEchoed++;
+}
+
+void updateEcnCountEchoed(
+    QuicConnectionStateBase& conn,
+    PacketNumberSpace pnSpace,
+    const ReadAckFrame& readAckFrame) {
+  // Track the reflected ECN markings in the current pn space
+  auto& ackState = getAckState(conn, pnSpace);
+  ackState.ecnECT0CountEchoed =
+      std::max(ackState.ecnECT0CountEchoed, readAckFrame.ecnECT0Count);
+  ackState.ecnECT1CountEchoed =
+      std::max(ackState.ecnECT1CountEchoed, readAckFrame.ecnECT1Count);
+  ackState.ecnCECountEchoed =
+      std::max(ackState.ecnCECountEchoed, readAckFrame.ecnCECount);
+}
 } // namespace quic
