@@ -406,17 +406,7 @@ class QuicStreamManager {
   /*
    * Clear all the currently open streams.
    */
-  void clearOpenStreams() {
-    openBidirectionalLocalStreams_.clear();
-    openUnidirectionalLocalStreams_.clear();
-    openBidirectionalPeerStreams_.clear();
-    openUnidirectionalPeerStreams_.clear();
-    openBidirectionalLocalStreamGroups_.clear();
-    openUnidirectionalLocalStreamGroups_.clear();
-    peerUnidirectionalStreamGroupsSeen_.clear();
-    peerBidirectionalStreamGroupsSeen_.clear();
-    streams_.clear();
-  }
+  void clearOpenStreams();
 
   /*
    * Return a const reference to the underlying container holding the stream
@@ -468,15 +458,6 @@ class QuicStreamManager {
    */
   bool setStreamPriority(StreamId id, Priority priority);
 
-  // TODO figure out a better interface here.
-  /*
-   * Returns a mutable reference to the container holding the writable stream
-   * IDs.
-   */
-  auto& writableStreams() {
-    return writableStreams_;
-  }
-
   auto& writableDSRStreams() {
     return writableDSRStreams_;
   }
@@ -837,12 +818,9 @@ class QuicStreamManager {
   /*
    * Consume the flow control updated streams using the parameter vector.
    */
-  auto consumeFlowControlUpdated(std::vector<StreamId>&& storage) {
-    std::vector<StreamId> result = storage;
-    result.clear();
-    result.reserve(flowControlUpdated_.size());
-    result.insert(
-        result.end(), flowControlUpdated_.begin(), flowControlUpdated_.end());
+  std::vector<StreamId> consumeFlowControlUpdated() {
+    std::vector<StreamId> result(
+        flowControlUpdated_.begin(), flowControlUpdated_.end());
     flowControlUpdated_.clear();
     return result;
   }
@@ -936,23 +914,24 @@ class QuicStreamManager {
   /*
    * Consume the new peer streams using the parameter vector.
    */
-  auto consumeNewPeerStreams(std::vector<StreamId>&& storage) {
-    return swapStreams(newPeerStreams_, std::move(storage));
+  std::vector<StreamId> consumeNewPeerStreams() {
+    std::vector<StreamId> res{std::move(newPeerStreams_)};
+    return res;
   }
 
   /*
    * Consume the new peer streams in groups using the parameter vector.
    */
-  auto consumeNewGroupedPeerStreams(std::vector<StreamId>&& storage) {
-    return swapStreams(newGroupedPeerStreams_, std::move(storage));
+  std::vector<StreamId> consumeNewGroupedPeerStreams() {
+    std::vector<StreamId> res{std::move(newGroupedPeerStreams_)};
+    return res;
   }
 
   /*
    * Consume the new peer stream groups using the parameter vector.
    */
   auto consumeNewPeerStreamGroups() {
-    decltype(newPeerStreamGroups_) result;
-    result.swap(newPeerStreamGroups_);
+    decltype(newPeerStreamGroups_) result{std::move(newPeerStreamGroups_)};
     return result;
   }
 
@@ -976,10 +955,8 @@ class QuicStreamManager {
    * Consume the stop sending streams.
    */
   auto consumeStopSending() {
-    std::vector<std::pair<StreamId, ApplicationErrorCode>> result;
-    result.reserve(stopSendingStreams_.size());
-    result.insert(
-        result.end(), stopSendingStreams_.begin(), stopSendingStreams_.end());
+    std::vector<std::pair<const StreamId, const ApplicationErrorCode>> result(
+        stopSendingStreams_.begin(), stopSendingStreams_.end());
     stopSendingStreams_.clear();
     return result;
   }
@@ -1105,20 +1082,6 @@ class QuicStreamManager {
       StreamGroupId& groupId,
       StreamIdSet& streamGroups);
 
-  /*
-   * Helper to consume new stream ids.
-   */
-  std::vector<StreamId> swapStreams(
-      std::vector<StreamId>& src,
-      std::vector<StreamId>&& dst) {
-    std::vector<StreamId> result = dst;
-    result.clear();
-    result.reserve(src.size());
-    result.insert(result.end(), src.begin(), src.end());
-    src.clear();
-    return result;
-  }
-
   QuicConnectionStateBase& conn_;
   QuicNodeType nodeType_;
 
