@@ -13,6 +13,7 @@
 #include <quic/codec/QuicWriteCodec.h>
 #include <quic/codec/Types.h>
 #include <quic/common/BufAccessor.h>
+#include <quic/common/CircularDeque.h>
 #include <quic/congestion_control/CongestionController.h>
 #include <quic/congestion_control/PacketProcessor.h>
 #include <quic/congestion_control/ThrottlingSignalProvider.h>
@@ -65,10 +66,15 @@ struct OutstandingsInfo {
   // declared lost, this counter will be decreased.
   uint64_t dsrCount{0};
 
+  // We just use this to get the correct number of outstanding packets. We
+  // subtract the number of packets declared lost and scheduled for destruction
+  // from the number of packets in the outstanding packets list.
+  uint64_t scheduledForDestructionCount{0};
+
   // Number of packets outstanding and not declared lost.
   uint64_t numOutstanding() {
-    CHECK_GE(packets.size(), declaredLostCount);
-    return packets.size() - declaredLostCount;
+    CHECK_GE(packets.size(), declaredLostCount + scheduledForDestructionCount);
+    return packets.size() - declaredLostCount - scheduledForDestructionCount;
   }
 
   // Total number of cloned packets.
@@ -267,6 +273,7 @@ using FrameList = std::vector<QuicSimpleFrame>;
 class CongestionControllerFactory;
 class LoopDetectorCallback;
 class PendingPathRateLimiter;
+class EcnL4sTracker;
 
 struct ReadDatagram {
   ReadDatagram(TimePoint recvTimePoint, BufQueue data)
@@ -348,6 +355,17 @@ struct QuicConnectionStateBase : public folly::DelayedDestruction {
 
   // Write cipher for 1-RTT data
   std::unique_ptr<Aead> oneRttWriteCipher;
+  ProtectionType oneRttWritePhase{ProtectionType::KeyPhaseZero};
+  uint64_t oneRttWritePacketsSentInCurrentPhase = 0;
+  // Whether there is a locally-initiated key update that is pending
+  // verification. This is used to decide whether the next variable should be
+  // updated or not.
+  bool oneRttWritePendingVerification{false};
+  // In a key update is pending, this holds the first packet number sent in the
+  // current (updated) phase. The peer must acknowledge this packet in the same
+  // phase, responding in a different phase is a protocol violation. Once the
+  // packet is acked, this value will be cleared.
+  folly::Optional<PacketNum> oneRttWritePendingVerificationPacketNumber;
 
   // Write cipher for packets with initial keys.
   std::unique_ptr<Aead> initialWriteCipher;
@@ -509,6 +527,15 @@ struct QuicConnectionStateBase : public folly::DelayedDestruction {
   // Current state of flow control.
   ConnectionFlowControlState flowControlState;
 
+  struct PendingWriteBatch {
+    std::unique_ptr<folly::IOBuf> buf;
+    // More fields will be needed here for other batch writer types.
+  };
+
+  // A write batch that was attempted but did not succeed.
+  // This is only used by the SinglePacketBackpressureBatchWriter.
+  PendingWriteBatch pendingWriteBatch_;
+
   // The outstanding path challenge
   folly::Optional<PathChallengeFrame> outstandingPathValidation;
 
@@ -642,9 +669,9 @@ struct QuicConnectionStateBase : public folly::DelayedDestruction {
     uint32_t maxReadBufferSize{kDefaultMaxDatagramsBuffered};
     uint32_t maxWriteBufferSize{kDefaultMaxDatagramsBuffered};
     // Buffers Incoming Datagrams
-    std::deque<ReadDatagram> readBuffer;
+    CircularDeque<ReadDatagram> readBuffer;
     // Buffers Outgoing Datagrams
-    std::deque<BufQueue> writeBuffer;
+    CircularDeque<BufQueue> writeBuffer;
   };
 
   DatagramState datagramState;
@@ -674,6 +701,17 @@ struct QuicConnectionStateBase : public folly::DelayedDestruction {
     uint64_t targetWriteCount;
   };
   SocketCmsgsState socketCmsgsState;
+
+  ECNState ecnState{ECNState::NotAttempted};
+  std::shared_ptr<EcnL4sTracker> ecnL4sTracker;
+  union TosHeader {
+    uint8_t value{0};
+    struct components {
+      unsigned int ecn : 2;
+      unsigned int dscp : 6;
+    } fields;
+  };
+  TosHeader socketTos;
 };
 
 std::ostream& operator<<(std::ostream& os, const QuicConnectionStateBase& st);
