@@ -102,7 +102,7 @@ struct QuicStreamLike {
 
   // List of bytes that have been read and buffered. We need to buffer
   // bytes in case we get bytes out of order.
-  std::deque<StreamBuffer> readBuffer;
+  CircularDeque<StreamBuffer> readBuffer;
 
   // List of bytes that have been written to the QUIC layer.
   BufQueue writeBuffer{};
@@ -119,13 +119,13 @@ struct QuicStreamLike {
   // offset ACKed. This allows us to track which delivery callbacks can be
   // called.
   template <class T>
-  using IntervalSetVec = SmallVec<T, 32>;
+  using IntervalSetVec = SmallVec<T, 3>;
   using AckedIntervals = IntervalSet<uint64_t, 1, IntervalSetVec>;
   AckedIntervals ackedIntervals;
 
   // Stores a list of buffers which have been marked as loss by loss detector.
   // Each one represents one StreamFrame that was written.
-  std::deque<StreamBuffer> lossBuffer;
+  CircularDeque<StreamBuffer> lossBuffer;
 
   // Current offset of the start bytes in the write buffer.
   // This changes when we pop stuff off the writeBuffer.
@@ -335,6 +335,11 @@ struct QuicStreamState : public QuicStreamLike {
     uint64_t peerAdvertisedMaxOffset{0};
     // Time at which the last flow control update was sent by the transport.
     folly::Optional<TimePoint> timeOfLastFlowControlUpdate;
+    // A flag indicating if the stream has a pending blocked frame to the peer
+    // (blocked frame sent, but a stream flow control update has not been
+    // received yet). Set when we write a blocked data frame on the stream;
+    // cleared when we receive a flow control update for the stream.
+    bool pendingBlockedFrame{false};
   };
 
   StreamFlowControlState flowControlState;
@@ -480,7 +485,7 @@ struct QuicStreamState : public QuicStreamLike {
   folly::F14FastMap<uint64_t, WriteBufferMeta> retransmissionBufMetas;
 
   // WriteBufferMetas that's already marked lost. They will be retransmitted.
-  std::deque<WriteBufferMeta> lossBufMetas;
+  CircularDeque<WriteBufferMeta> lossBufMetas;
 
   uint64_t streamLossCount{0};
 
